#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging

logger = logging.getLogger(__name__)

import enum
from typing import Callable, List

import bdkpython as bdk
from PyQt6.QtCore import QLocale, QObject, Qt, QTimer, pyqtSignal
from PyQt6.QtWidgets import QApplication, QLabel, QPushButton, QVBoxLayout

from bitcoin_safe.config import UserConfig
from bitcoin_safe.util import block_explorer_URL_of_projected_block, unit_fee_str

from ...html import html_f
from ...mempool import MempoolData, fee_to_color, mempoolFeeColors
from .invisible_scroll_area import InvisibleScrollArea
from .util import center_in_widget, open_website


def format_block_number(block_number) -> str:
    return QLocale().toString(int(block_number))


class BlockType(enum.Enum):
    projected = enum.auto()
    confirmed = enum.auto()
    unconfirmed = enum.auto()


class BaseBlockLabel(QLabel):
    def __init__(self, network: bdk.Network, text: str = "", parent=None) -> None:
        super().__init__(text, parent)
        self.network = network

        self.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.setWordWrap(True)
        self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)
        self.setHidden(not text)

    def setText(self, s: str | None) -> None:
        self.setHidden(not s)
        return super().setText(s)


class LabelTitle(BaseBlockLabel):
    def set(self, text: str, block_type: BlockType) -> None:
        self.setText(html_f(text, color="white" if block_type else "black", size="16px"))


class LabelApproximateMedianFee(BaseBlockLabel):
    def set(self, median_fee: float, block_type: BlockType) -> None:
        s = f"~{int(median_fee)} {unit_fee_str(self.network)}"

        self.setText(html_f(s, color="white" if block_type else "black", size="12px"))


class LabelExactMedianFee(BaseBlockLabel):
    def set(self, median_fee: float, block_type: BlockType) -> None:
        s = f"{round(median_fee, 1)} {unit_fee_str(self.network)}"

        self.setText(html_f(s, color="white" if block_type else "black", size="12px"))


class LabelNumberConfirmations(BaseBlockLabel):
    def set(self, i: int, block_type: BlockType) -> None:
        s = f"{i} Confirmation{'s' if i>1 else ''}"

        self.setText(html_f(s, color="white" if block_type else "black", size="12px"))


class LabelBlockHeight(BaseBlockLabel):
    def set(self, i: int, block_type: BlockType) -> None:
        s = f"{round(i)}. Block"

        self.setText(html_f(s, color="white" if block_type else "black", size="12px"))


class LabelFeeRange(BaseBlockLabel):
    def set(self, min_fee: float, max_fee: float) -> None:
        s = f"{int(min_fee)} - {int(max_fee)} {unit_fee_str(self.network)}"

        self.setText(html_f(s, color="#eee002", size="10px"))


class LabelTimeEstimation(BaseBlockLabel):
    def set(self, block_number: int, block_type: BlockType) -> None:
        if block_number < 6:
            s = self.tr("~in {t} min").format(t=(block_number) * 10)
        else:
            s = self.tr("~in {t} hours").format(t=round((block_number) / 6))

        self.setText(html_f(s, color="white" if block_type else "black", size="12px"))


class LabelExplorer(BaseBlockLabel):
    def set(self, block_type: BlockType) -> None:
        s = "visit<br>mempool.space"
        self.setText(html_f(s, color="white" if block_type else "black", size="10px"))


class BlockButton(QPushButton):
    def __init__(self, network: bdk.Network, size=100, parent=None) -> None:
        super().__init__(parent=parent)

        # Create labels for each text line

        self.label_approximate_median_fee = LabelApproximateMedianFee(network)
        self.label_exact_median_fee = LabelExactMedianFee(network)
        self.label_number_confirmations = LabelNumberConfirmations(network)
        self.label_block_height = LabelBlockHeight(network)
        self.label_fee_range = LabelFeeRange(network)
        self.label_title = LabelTitle(network)
        self.label_time_estimation = LabelTimeEstimation(network)
        self.label_explorer = LabelExplorer(network)

        # define the order:
        self.labels = [
            self.label_approximate_median_fee,
            self.label_exact_median_fee,
            self.label_number_confirmations,
            self.label_block_height,
            self.label_fee_range,
            self.label_title,
            self.label_time_estimation,
            self.label_explorer,
        ]

        layout = center_in_widget(self.labels, self, direction="v")
        layout.setContentsMargins(0, 0, 0, 0)  # Left, Top, Right, Bottom margins

        # Ensure buttons are square
        self.setMinimumHeight(size)
        self.setMinimumWidth(size)

    def clear_labels(self) -> None:
        for label in self.labels:
            label.setText("")

    def _set_background_gradient(self, color_top: str, color_bottom: str) -> None:
        # Set the stylesheet for the QPushButton
        self.setStyleSheet(
            f"""
            QPushButton {{
                background-color: qlineargradient(spread:pad, x1:0, y1:0, x2:0, y2:1,
                                    stop:0 {color_bottom},
                                    stop:1 {color_top});
                color: white; /* Change this to set the text color */
            }}
        """
        )

    def set_background_gradient(self, min_fee: float, max_fee: float, block_type: BlockType) -> None:
        self.block_type = block_type
        if self.block_type == BlockType.confirmed:
            self._set_background_gradient("#115fb0", "#9239f3")
        else:
            self._set_background_gradient(
                fee_to_color(min_fee, mempoolFeeColors),
                fee_to_color(max_fee, mempoolFeeColors),
            )


class VerticalButtonGroup(InvisibleScrollArea):
    signal_button_click = pyqtSignal(int)

    def __init__(self, network: bdk.Network, button_count=3, parent=None, size=100) -> None:
        super().__init__(parent)
        layout = QVBoxLayout(self.content_widget)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setMinimumWidth(size + 30)
        # if button_count > 1:
        #     self.setMinimumHeight(size + 20)

        self.setWidgetResizable(True)
        self.buttons: List[BlockButton] = []

        # Create buttons
        for i in range(button_count):
            button = BlockButton(network=network, size=size)

            def create_signal_handler(index) -> Callable:
                def send_signal() -> None:
                    return self.signal_button_click.emit(index)

                return send_signal

            button.clicked.connect(create_signal_handler(i))

            self.buttons.append(button)
            layout.addWidget(button)
            layout.setAlignment(button, Qt.AlignmentFlag.AlignCenter)


class ObjectRequiringMempool(QObject):
    def __init__(self, mempool_data: MempoolData, parent=None) -> None:
        super().__init__(parent=parent)

        self.mempool_data = mempool_data

        self.timer = QTimer()
        self.timer.timeout.connect(self.mempool_data.set_data_from_mempoolspace)
        self.timer.start(10 * 60 * 1000)  # 10 minutes in milliseconds

    def set_mempool_block_unknown_fee_rate(self, i, confirmation_time: bdk.BlockTime | None = None) -> None:
        logger.error("This should not be called")


class BaseBlock(QObject):
    signal_click = pyqtSignal(float)

    def __init__(
        self,
        mempool_data: MempoolData,
        button_group: VerticalButtonGroup,
        confirmation_time: bdk.BlockTime | None = None,
        parent=None,
    ) -> None:
        QObject.__init__(self, parent=parent)
        self.confirmation_time = confirmation_time
        self.mempool_data = mempool_data
        self.button_group = button_group

        # signals
        self.button_group.signal_button_click.connect(self._on_button_click)
        self.mempool_data.signal_data_updated.connect(self.refresh)

    def refresh(self, **kwargs) -> None:
        pass

    def set_url(self, url: str) -> None:
        pass

    def _on_button_click(self, i: int) -> None:
        pass


class MempoolButtons(BaseBlock, ObjectRequiringMempool):
    "Showing multiple buttons of the next, the 2. and the 3. block templates according to the mempool"

    def __init__(self, mempool_data: MempoolData, max_button_count=3, parent=None) -> None:
        button_group = VerticalButtonGroup(
            network=mempool_data.network_config.network, button_count=max_button_count, parent=parent
        )
        BaseBlock.__init__(
            self, mempool_data=mempool_data, button_group=button_group, confirmation_time=None, parent=parent
        )
        ObjectRequiringMempool.__init__(self, mempool_data=mempool_data, parent=parent)

        self.refresh()

    def refresh(self, **kwargs) -> None:
        for i, button in enumerate(self.button_group.buttons):
            block_number = i + 1
            button.setVisible(i < max(1, self.mempool_data.num_mempool_blocks()))
            button.label_title.set(
                (
                    self.tr("Next Block")
                    if block_number == 1
                    else self.tr("{n}. Block").format(n=format_block_number(block_number))
                ),
                block_type=BlockType.projected,
            )
            button.label_time_estimation.set(block_number, block_type=BlockType.projected)
            button.label_approximate_median_fee.set(
                self.mempool_data.median_block_fee_rate(i), block_type=BlockType.projected
            )
            button.label_fee_range.set(*self.mempool_data.fee_rates_min_max(i))
            button.set_background_gradient(*self.mempool_data.fee_rates_min_max(i), BlockType.projected)

    def _on_button_click(self, i: int) -> None:
        logger.debug(f"Clicked button {i}: {self.mempool_data.median_block_fee_rate(i)}")
        self.signal_click.emit(self.mempool_data.median_block_fee_rate(i))


class MempoolProjectedBlock(BaseBlock, ObjectRequiringMempool):
    "The Button showing the block in which the fee_rate fits"

    def __init__(
        self,
        mempool_data: MempoolData,
        config: UserConfig,
        fee_rate: float = 1,
        parent=None,
    ) -> None:
        button_group = VerticalButtonGroup(
            network=mempool_data.network_config.network, size=100, button_count=1, parent=parent
        )

        BaseBlock.__init__(
            self, mempool_data=mempool_data, button_group=button_group, confirmation_time=None, parent=parent
        )
        ObjectRequiringMempool.__init__(self, mempool_data=mempool_data, parent=parent)

        self.config = config
        self.fee_rate = fee_rate
        self.url = ""

        self.refresh()

    def set_url(self, url: str) -> None:
        self.url = url

    def set_unknown_fee_rate(self) -> None:
        for button in self.button_group.buttons:
            button.clear_labels()
            button.label_title.set(self.tr("Unconfirmed"), BlockType.projected)
            button.label_explorer.set(BlockType.projected)
            button.set_background_gradient(0, 1, BlockType.projected)

    def refresh(self, fee_rate=None, **kwargs) -> None:
        self.fee_rate = fee_rate if fee_rate is not None else self.fee_rate
        # if self.fee_rate is None:
        #     self.set_unknown_fee_rate()
        #     return

        block_index = self.mempool_data.fee_rate_to_projected_block_index(self.fee_rate)

        for button in self.button_group.buttons:

            button.label_title.set(
                self.tr("~{n}. Block").format(n=format_block_number(block_index + 1)), BlockType.projected
            )
            button.label_approximate_median_fee.set(
                self.mempool_data.median_block_fee_rate(block_index), block_type=BlockType.projected
            )
            button.label_fee_range.set(*self.mempool_data.fee_rates_min_max(block_index))
            button.label_time_estimation.set(block_index + 1, BlockType.projected)
            button.set_background_gradient(
                *self.mempool_data.fee_rates_min_max(block_index), BlockType.projected
            )

    def _on_button_click(self, i: int) -> None:
        block_index = self.mempool_data.fee_rate_to_projected_block_index(self.fee_rate)
        url = (
            self.url
            if self.url
            else block_explorer_URL_of_projected_block(self.config.network_config.mempool_url, block_index)
        )
        if url:
            open_website(url)


class ConfirmedBlock(BaseBlock):
    "Showing a confirmed block"

    def __init__(
        self,
        mempool_data: MempoolData,
        url: str | None = None,
        confirmation_time: bdk.BlockTime | None = None,
        fee_rate: float | None = None,
        parent=None,
    ) -> None:
        button_group = VerticalButtonGroup(
            network=mempool_data.network_config.network, button_count=1, parent=parent, size=120
        )

        super().__init__(
            parent=parent,
            mempool_data=mempool_data,
            button_group=button_group,
            confirmation_time=confirmation_time,
        )

        self.fee_rate = fee_rate
        self.url = url

    def set_url(self, url: str) -> None:
        self.url = url

    def refresh(
        self, fee_rate=None, confirmation_time=None, chain_height: int | None = None, **kwargs
    ) -> None:
        self.fee_rate = fee_rate if fee_rate is not None else self.fee_rate
        self.confirmation_time = confirmation_time if confirmation_time else self.confirmation_time
        if not self.confirmation_time:
            return

        for i, button in enumerate(self.button_group.buttons):
            button.label_title.set(
                self.tr("Block {n}").format(n=format_block_number(self.confirmation_time.height)),
                BlockType.confirmed,
            )
            if chain_height is None:
                button.label_number_confirmations.setText("")
                button.label_block_height.set(self.confirmation_time.height, BlockType.confirmed)
            else:
                button.label_number_confirmations.set(
                    chain_height - self.confirmation_time.height + 1,
                    BlockType.confirmed,
                )
                button.label_block_height.setText("")

            button.label_explorer.set(BlockType.confirmed)
            if self.fee_rate:
                button.set_background_gradient(self.fee_rate, self.fee_rate, BlockType.confirmed)
                button.label_exact_median_fee.set(self.fee_rate, BlockType.confirmed)
            else:
                button.set_background_gradient(0, 1, BlockType.confirmed)
                button.label_exact_median_fee.setText("")

    def _on_button_click(self, i: int) -> None:
        if self.url:
            open_website(self.url)
        self.signal_click.emit(self.fee_rate)


if __name__ == "__main__":
    import sys

    from PyQt6.QtWidgets import QApplication

    app = QApplication(sys.argv)

    widget = VerticalButtonGroup(network=bdk.Network.REGTEST)
    widget.show()

    sys.exit(app.exec())



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
import typing

from bitcoin_qr_tools.qr_widgets import QRCodeWidgetSVG

from bitcoin_safe.config import UserConfig
from bitcoin_safe.descriptors import MultipathDescriptor
from bitcoin_safe.gui.qt.buttonedit import ButtonEdit
from bitcoin_safe.gui.qt.recipients import RecipientTabWidget
from bitcoin_safe.gui.qt.register_multisig import USBValidateAddressWidget
from bitcoin_safe.keystore import KeyStoreImporterTypes
from bitcoin_safe.mempool import MempoolData
from bitcoin_safe.util import serialized_to_hex

logger = logging.getLogger(__name__)

import bdkpython as bdk
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QKeyEvent
from PyQt6.QtWidgets import QFormLayout, QHBoxLayout, QSizePolicy, QVBoxLayout, QWidget

from ...signals import Signals
from ...wallet import Wallet
from .hist_list import HistList
from .util import Buttons, CloseButton, read_QIcon


class AddressDetailsAdvanced(QWidget):
    def __init__(
        self, bdk_address: bdk.Address, address_path_str: str, parent: typing.Optional["QWidget"]
    ) -> None:
        super().__init__(parent)

        form_layout = QFormLayout(self)
        form_layout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow)

        try:
            script_pubkey = serialized_to_hex(bdk_address.script_pubkey().to_bytes())
        except BaseException:
            script_pubkey = None
        if script_pubkey:
            pubkey_e = ButtonEdit(script_pubkey)
            pubkey_e.button_container.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
            pubkey_e.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)

            pubkey_e.add_copy_button()
            pubkey_e.setReadOnly(True)

            form_layout.addRow(self.tr("Script Pubkey"), pubkey_e)

        if address_path_str:
            der_path_e = ButtonEdit(address_path_str)
            der_path_e.add_copy_button()
            der_path_e.setFixedHeight(50)
            der_path_e.setReadOnly(True)

            form_layout.addRow(self.tr("Address descriptor"), der_path_e)


class AddressValidateTab(QWidget):
    def __init__(
        self,
        bdk_address: bdk.Address,
        wallet_descriptor: MultipathDescriptor,
        kind: bdk.KeychainKind,
        address_index: int,
        network: bdk.Network,
        signals: Signals,
        parent: typing.Optional["QWidget"],
    ) -> None:
        super().__init__(parent)

        self._layout = QHBoxLayout(self)

        edit_addr_descriptor = USBValidateAddressWidget(network=network, signals=signals)
        edit_addr_descriptor.set_descriptor(
            descriptor=wallet_descriptor,
            expected_address=bdk_address.as_string(),
            kind=kind,
            address_index=address_index,
        )
        self._layout.addWidget(edit_addr_descriptor)


class QRAddress(QRCodeWidgetSVG):
    def __init__(
        self,
    ) -> None:
        super().__init__(clickable=False)
        self.setMaximumSize(150, 150)

    def set_address(self, bdk_address: bdk.Address):
        self.set_data_list([bdk_address.to_qr_uri()])


class AddressDialog(QWidget):
    def __init__(
        self,
        fx,
        config: UserConfig,
        signals: Signals,
        wallet: Wallet,
        address: str,
        mempool_data: MempoolData,
        parent=None,
    ) -> None:
        super().__init__(parent, Qt.WindowType.Window)
        self.setWindowTitle(self.tr("Address"))
        self.mempool_data = mempool_data
        self.address = address
        self.bdk_address = bdk.Address(address, network=wallet.network)
        self.fx = fx
        self.config = config
        self.wallet: Wallet = wallet
        self.signals = signals
        self.saved = True

        self.setMinimumWidth(700)
        vbox = QVBoxLayout()
        self.setLayout(vbox)

        upper_widget = QWidget()
        # upper_widget.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self.upper_widget_layout = QHBoxLayout(upper_widget)
        self.upper_widget_layout.setContentsMargins(0, 0, 0, 0)

        vbox.addWidget(upper_widget)

        self.recipient_tabs = RecipientTabWidget(
            network=wallet.network,
            allow_edit=False,
            parent=self,
            signals=self.signals,
            tab_string=self.tr('Address of wallet "{id}"'),
        )
        self.recipient_tabs.address = self.address
        label = wallet.labels.get_label(self.address)
        self.recipient_tabs.label = label if label else ""
        self.recipient_tabs.amount = wallet.get_addr_balance(self.address).total

        self.upper_widget_layout.addWidget(self.recipient_tabs)

        self.tab_advanced = AddressDetailsAdvanced(
            bdk_address=self.bdk_address,
            address_path_str=self.wallet.get_address_path_str(address),
            parent=self,
        )
        self.recipient_tabs.addTab(self.tab_advanced, "")

        address_info = self.wallet.get_address_info_min(address)
        if address_info:
            self.tab_validate = AddressValidateTab(
                bdk_address=self.bdk_address,
                network=config.network,
                signals=self.signals,
                wallet_descriptor=self.wallet.multipath_descriptor,
                kind=address_info.keychain,
                address_index=address_info.index,
                parent=self,
            )
            self.recipient_tabs.addTab(
                self.tab_validate, read_QIcon(KeyStoreImporterTypes.hwi.icon_filename), ""
            )

        self.qr_code = QRAddress()
        self.qr_code.set_address(self.bdk_address)
        self.upper_widget_layout.addWidget(self.qr_code)

        self.hist_list = HistList(
            self.fx,
            self.config,
            self.signals,
            self.mempool_data,
            self.wallet.id,
            hidden_columns=[
                HistList.Columns.TXID,
                HistList.Columns.BALANCE,
            ],
            address_domain=[self.address],
            column_widths={HistList.Columns.TXID: 100},
        )
        vbox.addWidget(self.hist_list)

        vbox.addLayout(Buttons(CloseButton(self)))
        self.setupUi()

    # Override keyPressEvent method
    def keyPressEvent(self, event: QKeyEvent) -> None:  # type: ignore[override]
        # Check if the pressed key is 'Esc'
        if event.key() == Qt.Key.Key_Escape:
            # Close the widget
            self.close()

    def setupUi(self) -> None:
        self.recipient_tabs.updateUi()
        self.recipient_tabs.setTabText(self.recipient_tabs.indexOf(self.tab_advanced), self.tr("Advanced"))
        self.recipient_tabs.setTabText(self.recipient_tabs.indexOf(self.tab_validate), self.tr("Validate"))



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from typing import List, Optional

from bitcoin_safe.descriptors import MultipathDescriptor
from bitcoin_safe.gui.qt.address_edit import AddressEdit
from bitcoin_safe.gui.qt.analyzer_indicator import ElidedLabel
from bitcoin_safe.gui.qt.tutorial_screenshots import ScreenshotsRegisterMultisig
from bitcoin_safe.keystore import KeyStore, KeyStoreImporterTypes
from bitcoin_safe.signals import Signals

logger = logging.getLogger(__name__)


import bdkpython as bdk
from bitcoin_usb.gui import USBGui
from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from ...signals import Signals
from .util import Message, MessageType, generate_help_button, read_QIcon


class USBValidateAddressWidget(QWidget):
    def __init__(
        self,
        network: bdk.Network,
        signals: Signals,
    ) -> None:
        super().__init__()
        self.signals = signals
        self.network = network
        self.descriptor: Optional[MultipathDescriptor] = None
        self.expected_address = ""
        self.address_index = 0
        self.kind = bdk.KeychainKind.EXTERNAL
        self.usb = USBGui(self.network, allow_emulators_only_for_testnet_works=True)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        self._layout = QVBoxLayout(self)
        self._layout.setAlignment(Qt.AlignmentFlag.AlignVCenter)

        self.label_expected_address = QLabel()
        self._layout.addWidget(self.label_expected_address)

        self.edit_address = AddressEdit(network=network, allow_edit=False, parent=self, signals=self.signals)
        self._layout.addWidget(self.edit_address)

        # Create buttons and layout
        self.button_box = QDialogButtonBox()
        self._layout.addWidget(self.button_box)
        self._layout.setAlignment(self.button_box, Qt.AlignmentFlag.AlignCenter)

        self.button_validate_address = QPushButton()
        self.button_validate_address.setIcon(read_QIcon(KeyStoreImporterTypes.hwi.icon_filename))
        self.button_validate_address.clicked.connect(self.on_button_click)
        self.button_box.addButton(self.button_validate_address, QDialogButtonBox.ButtonRole.AcceptRole)

        self.updateUi()
        self.signals.language_switch.connect(self.updateUi)

    def updateUi(self) -> None:
        self.button_validate_address.setText(self.tr("Validate address"))
        self.label_expected_address.setText(self.tr("Validate receive address:"))

    def set_descriptor(
        self,
        descriptor: MultipathDescriptor,
        expected_address: str,
        kind: bdk.KeychainKind = bdk.KeychainKind.EXTERNAL,
        address_index: int = 0,
    ) -> None:
        self.descriptor = descriptor
        self.expected_address = expected_address
        self.kind = kind
        self.address_index = address_index
        self.edit_address.setText(self.expected_address)

        self.updateUi()

    def on_button_click(
        self,
    ) -> bool:
        if not self.descriptor:
            logger.error("descriptor not set")
            return False

        address_descriptor = self.descriptor.address_descriptor(
            kind=self.kind, address_index=self.address_index
        )
        try:
            address = self.usb.display_address(address_descriptor)
        except Exception as e:
            Message(str(e), type=MessageType.Error)
            return False

        return bool(address)


class USBRegisterMultisigWidget(USBValidateAddressWidget):
    def __init__(self, network: bdk.Network, signals: Signals) -> None:
        screenshots = ScreenshotsRegisterMultisig()
        self.button_help = generate_help_button(screenshots, title="Help")

        super().__init__(network, signals)

        self.button_box.addButton(self.button_help, QDialogButtonBox.ButtonRole.HelpRole)

        self.xpubs_widget = QWidget()
        self.xpubs_widget_layout = QHBoxLayout(self.xpubs_widget)
        self.label_title_keystore = QLabel()
        self.label_xpubs_keystore = ElidedLabel(elide_mode=Qt.TextElideMode.ElideMiddle)
        self.xpubs_widget_layout.addWidget(self.label_title_keystore)
        self.xpubs_widget_layout.addWidget(self.label_xpubs_keystore)

        self._layout.insertWidget(0, self.xpubs_widget)

    def updateUi(self) -> None:
        super().updateUi()
        self.setWindowTitle(self.tr("Register Multisig wallet on hardware signer"))
        self.button_validate_address.setText(self.tr("Register Multisig"))
        self.button_help.setText(self.tr("Help"))

    def on_button_click(
        self,
    ) -> bool:
        result = super().on_button_click()

        if result:
            self.close()
            Message(
                self.tr("Successfully registered multisig wallet on hardware signer"),
                type=MessageType.Info,
                icon=read_QIcon("checkmark.svg"),
            )
        return result

    def set_descriptor(  # type: ignore
        self,
        keystores: List[KeyStore],
        descriptor: MultipathDescriptor,
        expected_address: str,
        kind: bdk.KeychainKind = bdk.KeychainKind.EXTERNAL,
        address_index: int = 0,
    ) -> None:
        super().set_descriptor(
            descriptor=descriptor, expected_address=expected_address, kind=kind, address_index=address_index
        )

        text_titles = "\n".join([f"{keystore.label}:" for keystore in keystores])
        text_xpubs = "\n".join([keystore.xpub for keystore in keystores])
        self.label_title_keystore.setText(text_titles)
        self.label_xpubs_keystore.setText(text_xpubs)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from typing import Tuple

import bdkpython as bdk
from PyQt6 import QtGui, QtWidgets

from ...util import Satoshis


class BTCSpinBox(QtWidgets.QDoubleSpinBox):
    "A Satoshi Spin Box.  The value stored is in Satoshis."

    def __init__(self, network: bdk.Network, parent=None) -> None:
        super().__init__(parent)
        self.network = network
        self._is_max = False
        self.setDecimals(0)  # Set the number of decimal places
        self.setRange(0, 21e6 * 1e8)  # Define range as required

    def set_max(self, value: bool) -> None:
        self.setDisabled(value)
        self._is_max = value
        self.setValue(super().value())

    def value(self) -> int:
        return round(super().value())

    def textFromValue(self, value: int) -> str:  # type: ignore[override]
        if self._is_max:
            return self.tr("Max ≈ {amount}").format(amount=str(Satoshis(value, self.network)))
        return str(Satoshis(value, self.network))

    def valueFromText(self, text: str | None) -> int:
        if self._is_max:
            return 0
        return Satoshis(text if text else 0, self.network).value

    def validate(self, text: str | None, pos: int) -> Tuple[QtGui.QValidator.State, str, int]:
        if text is None:
            text = ""
        try:
            # Try to convert the text to a float
            self.valueFromText(text)
            # If it succeeds, the text is valid
            return QtGui.QValidator.State.Acceptable, text, pos
        except ValueError:
            # If it fails, the text is not valid
            return QtGui.QValidator.State.Invalid, text, pos



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from pathlib import Path

from packaging import version

from bitcoin_safe.gui.qt.unique_deque import UniqueDeque

from .execute_config import DEFAULT_MAINNET

logger = logging.getLogger(__name__)

import os
from typing import Any, Dict, List, Optional

import appdirs
import bdkpython as bdk

from .network_config import NetworkConfig, NetworkConfigs
from .storage import BaseSaveableClass
from .util import (
    briefcase_project_dir,
    path_to_rel_home_path,
    rel_home_path_to_abs_path,
)

MIN_RELAY_FEE = 1
FEE_RATIO_HIGH_WARNING = 0.05  # warn user if fee/amount for on-chain tx is higher than this
NO_FEE_WARNING_BELOW = 10  # sat/vB


RECENT_WALLET_MAXLEN = 15


class UserConfig(BaseSaveableClass):
    known_classes = {**BaseSaveableClass.known_classes, "NetworkConfigs": NetworkConfigs}
    VERSION = "0.1.6"

    app_name = "bitcoin_safe"
    locales_path = briefcase_project_dir() / "gui" / "locales"
    config_dir = Path(appdirs.user_config_dir(app_name))
    config_file = config_dir / (app_name + ".conf")

    fee_ranges = {
        bdk.Network.BITCOIN: [1.0, 1000],
        bdk.Network.REGTEST: [0.0, 1000],
        bdk.Network.SIGNET: [0.0, 1000],
        bdk.Network.TESTNET: [0.0, 1000],
    }

    def __init__(self) -> None:
        self.network_configs = NetworkConfigs()
        self.network: bdk.Network = bdk.Network.BITCOIN if DEFAULT_MAINNET else bdk.Network.TESTNET
        self.last_wallet_files: Dict[str, List[str]] = {}  # network:[file_path0]
        self.opened_txlike: Dict[str, List[str]] = {}  # network:[serializedtx, serialized psbt]
        self.data_dir = appdirs.user_data_dir(self.app_name)
        self.is_maximized = False
        self.recently_open_wallets: Dict[bdk.Network, UniqueDeque[str]] = {
            network: UniqueDeque(maxlen=RECENT_WALLET_MAXLEN) for network in bdk.Network
        }
        self.language_code: Optional[str] = None

    def add_recently_open_wallet(self, file_path: str) -> None:
        self.recently_open_wallets[self.network].append(file_path)

    @property
    def network_config(self) -> NetworkConfig:
        return self.network_configs.configs[self.network.name]

    @property
    def wallet_dir(self) -> str:
        return os.path.join(self.config_dir, self.network.name)

    def get(self, key: str, default=None) -> Any:
        "For legacy reasons"
        if hasattr(self, key):
            return getattr(self, key)
        else:
            return default

    def dump(self) -> Dict[str, Any]:
        d = super().dump()
        d.update(self.__dict__.copy())

        # for better portability between computers we make this relative to the home folder
        d["data_dir"] = str(path_to_rel_home_path(self.data_dir))

        d["recently_open_wallets"] = {
            network.name: list(v) for network, v in self.recently_open_wallets.items()
        }
        return d

    @classmethod
    def from_dump(cls, dct: Dict, class_kwargs=None) -> "UserConfig":
        super()._from_dump(dct, class_kwargs=class_kwargs)
        dct["recently_open_wallets"] = {
            bdk.Network._member_map_[k]: UniqueDeque(v, maxlen=RECENT_WALLET_MAXLEN)
            for k, v in dct.get(
                "recently_open_wallets",
                {network.name: UniqueDeque(maxlen=RECENT_WALLET_MAXLEN) for network in bdk.Network},
            ).items()
        }
        # for better portability between computers the saved string is relative to the home folder
        dct["data_dir"] = rel_home_path_to_abs_path(dct["data_dir"])
        # dct["config_dir"] = rel_home_path_to_abs_path(dct["config_dir"])
        # dct["config_file"] = rel_home_path_to_abs_path(dct["config_file"])

        u = cls()

        for k, v in dct.items():
            if v is not None:  # only overwrite the default value, if there is a value
                setattr(u, k, v)
        return u

    @classmethod
    def from_dump_migration(cls, dct: Dict[str, Any]) -> Dict[str, Any]:
        "this class should be overwritten in child classes"
        if version.parse(str(dct["VERSION"])) <= version.parse("0.1.0"):
            network_config: NetworkConfig = dct["network_config"]
            dct["network_configs"] = {network.name: NetworkConfig(network=network) for network in bdk.Network}
            dct["network_configs"][network_config.network.name] = network_config
            dct["network"] = network_config.network
            del dct["network_config"]
        if version.parse(str(dct["VERSION"])) <= version.parse("0.1.1"):
            if "enable_opportunistic_merging_fee_rate" in dct:
                del dct["enable_opportunistic_merging_fee_rate"]
        if version.parse(str(dct["VERSION"])) <= version.parse("0.1.2"):
            if "network_configs" in dct:
                del dct["network_configs"]
        if version.parse(str(dct["VERSION"])) <= version.parse("0.1.3"):
            if "recently_open_wallets" in dct:
                del dct["recently_open_wallets"]
        if version.parse(str(dct["VERSION"])) <= version.parse("0.1.4"):
            if "recently_open_wallets" in dct:
                del dct["recently_open_wallets"]
        if version.parse(str(dct["VERSION"])) <= version.parse("0.1.6"):
            if "config_dir" in dct:
                del dct["config_dir"]
            if "config_file" in dct:
                del dct["config_file"]

        # now the VERSION is newest, so it can be deleted from the dict
        if "VERSION" in dct:
            del dct["VERSION"]
        return dct

    @classmethod
    def exists(cls, password=None, file_path=None) -> bool:
        if file_path is None:
            file_path = cls.config_file
        return os.path.isfile(file_path)

    @classmethod
    def from_file(cls, password: str | None = None, file_path: Path | None = None) -> "UserConfig":
        if file_path is None:
            file_path = cls.config_file
        if os.path.isfile(file_path):
            return super()._from_file(str(file_path), password=password)
        else:
            return UserConfig()

    def save(self) -> None:  # type: ignore
        super().save(self.config_file)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from typing import Callable, Optional

import bdkpython as bdk
from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtGui import QDragEnterEvent, QDropEvent, QKeyEvent, QKeySequence, QShortcut
from PyQt6.QtWidgets import (
    QApplication,
    QDialog,
    QDialogButtonBox,
    QLabel,
    QVBoxLayout,
    QWidget,
)

from bitcoin_safe.gui.qt.buttonedit import ButtonEdit
from bitcoin_safe.gui.qt.custom_edits import AnalyzerTextEdit
from bitcoin_safe.i18n import translate

logger = logging.getLogger(__name__)


def is_binary(file_path) -> bool:
    """Check if a file is binary or text.

    Returns True if binary, False if text.
    """
    try:
        with open(file_path, "r") as f:
            for chunk in iter(lambda: f.read(1024), ""):
                if "\0" in chunk:  # found null byte
                    return True
    except UnicodeDecodeError:
        return True

    return False


def file_to_str(file_path) -> str:
    if is_binary(file_path):
        with open(file_path, "rb") as f:
            return bytes(f.read()).hex()
    else:
        with open(file_path, "r") as f:
            return f.read()


class DragAndDropTextEdit(AnalyzerTextEdit):
    def __init__(
        self,
        parent=None,
        callback_enter=None,
        callback_esc=None,
        process_filepath: Optional[Callable[[str], None]] = None,
    ) -> None:
        super().__init__(parent=parent)
        self.process_filepath = process_filepath
        self.callback_enter = callback_enter
        self.callback_esc = callback_esc

    def keyPressEvent(self, event: QKeyEvent | None) -> None:
        if not event:
            super().keyPressEvent(event)
            return

        if event.key() == Qt.Key.Key_Return or event.key() == Qt.Key.Key_Enter:
            if self.callback_enter:
                self.callback_enter(self.toPlainText())
        elif event.key() == Qt.Key.Key_Escape:
            if self.callback_esc:
                self.callback_esc()
        super().keyPressEvent(event)

    def dragEnterEvent(self, event: QDragEnterEvent | None) -> None:
        if not event:
            super().dragEnterEvent(event)
            return

        mime_data = event.mimeData()
        if mime_data and mime_data.hasUrls():
            event.accept()
        else:
            event.ignore()

        super().dragEnterEvent(event)

    def dropEvent(self, event: QDropEvent | None) -> None:
        if not event:
            super().dropEvent(event)
            return

        mime_data = event.mimeData()
        if mime_data:
            file_path = mime_data.urls()[0].toLocalFile()
            if self.process_filepath:
                self.process_filepath(file_path)

        super().dropEvent(event)


class DragAndDropButtonEdit(ButtonEdit):
    signal_drop_file = pyqtSignal(str)

    def __init__(
        self,
        network: bdk.Network,
        parent=None,
        callback_enter=None,
        callback_esc=None,
        file_filter=translate("DragAndDropButtonEdit", "All Files (*);;PSBT (*.psbt);;Transation (*.tx)"),
    ) -> None:
        super().__init__(
            parent=parent,
            input_field=DragAndDropTextEdit(
                parent=parent,
                callback_enter=callback_enter,
                callback_esc=callback_esc,
                process_filepath=self.process_filepath,
            ),
        )
        self.network = network

        self.add_qr_input_from_camera_button(
            network=self.network,
        )
        self.button_open_file = self.add_open_file_button(self.process_filepath, filter=file_filter)

    def process_filepath(self, file_path: str) -> None:
        s = file_to_str(file_path)
        self.setText(s)
        self.signal_drop_file.emit(s)


class ImportDialog(QDialog):
    def __init__(
        self,
        network: bdk.Network,
        window_title="Open Transaction or PSBT",
        on_open=None,
        parent=None,
        text_button_ok="OK",
        text_instruction_label="Please paste your Bitcoin Transaction or PSBT in here, or drop a file",
        instruction_widget: Optional[QWidget] = None,
        text_placeholder="Paste your Bitcoin Transaction or PSBT in here or drop a file",
    ) -> None:
        super().__init__(parent)
        self.on_open = on_open

        self.setWindowTitle(window_title)
        layout = QVBoxLayout()

        self.instruction_label = QLabel(text_instruction_label)
        self.text_edit = DragAndDropButtonEdit(
            network=network,
            callback_enter=self.process_input,
            callback_esc=self.close,
        )
        self.text_edit.setPlaceholderText(text_placeholder)

        if instruction_widget:
            layout.addWidget(instruction_widget)
        layout.addWidget(self.instruction_label)
        layout.addWidget(self.text_edit)

        self.setLayout(layout)

        # buttons
        self.buttonBox = QDialogButtonBox(self)
        self.cancel_button = self.buttonBox.addButton(QDialogButtonBox.StandardButton.Cancel)
        if self.cancel_button:
            self.cancel_button.clicked.connect(self.close)
        # self.button_file = self.buttonBox.addButton(QDialogButtonBox.Open)
        self.button_ok = self.buttonBox.addButton(QDialogButtonBox.StandardButton.Ok)
        if self.button_ok:
            self.button_ok.setDefault(True)
            self.button_ok.setText(text_button_ok)
            self.button_ok.clicked.connect(lambda: self.process_input(self.text_edit.text()))

        layout.addWidget(self.buttonBox)

        # connect signals
        self.text_edit.signal_drop_file.connect(self.process_input)

        shortcut = QShortcut(QKeySequence("Return"), self)
        shortcut.activated.connect(self.process_input)

    def keyPressEvent(self, event: QKeyEvent | None) -> None:
        if event and event.key() == Qt.Key.Key_Escape:
            self.close()

    def process_input(self, s: str) -> None:
        if self.on_open:
            self.on_open(s)

        # close lets the entire application crash
        self.deleteLater()


if __name__ == "__main__":
    import sys

    from PyQt6.QtWidgets import QApplication

    app = QApplication(sys.argv)

    dialog = ImportDialog(network=bdk.Network.REGTEST, on_open=print)
    dialog.show()

    sys.exit(app.exec())



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging

from bitcoin_safe.gui.qt.analyzers import AnalyzerMessage, AnalyzerState, BaseAnalyzer
from bitcoin_safe.gui.qt.custom_edits import AnalyzerLineEdit, AnalyzerTextEdit

logger = logging.getLogger(__name__)

from typing import List, Optional, Union

from PyQt6.QtCore import QSize, Qt
from PyQt6.QtGui import QFontMetrics, QPainter, QPixmap
from PyQt6.QtWidgets import (
    QApplication,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QSizePolicy,
    QSpacerItem,
    QStyle,
    QWidget,
)


class ElidedLabel(QLabel):
    def __init__(self, elide_mode: Qt.TextElideMode = Qt.TextElideMode.ElideRight):
        super().__init__()
        self.elide_mode = elide_mode

    def paintEvent(self, event):
        painter = QPainter(self)
        metrics = QFontMetrics(self.font())
        elided = "\n".join(
            [metrics.elidedText(line, self.elide_mode, self.width()) for line in self.text().split("\n")]
        )
        painter.drawText(self.rect(), self.alignment(), elided)

    def _requirey_y_size(self) -> int:
        # Create a QFontMetrics object to measure text dimensions
        metrics = QFontMetrics(self.font())
        text_height = metrics.height()  # Height of one line of text

        # Count the number of lines in the label's text
        number_of_lines = len(self.text().split("\n"))

        # Calculate the total height based on the number of lines
        total_height = number_of_lines * text_height
        return total_height

    def minimumSizeHint(self):
        return QSize(1, self._requirey_y_size())  # Fixed width, dynamic height


class AnalyzerIndicator(QWidget):
    def __init__(
        self,
        line_edits: List[Union[AnalyzerLineEdit, AnalyzerTextEdit]],
        icon_OK: Optional[QPixmap] = None,
        icon_warning: Optional[QPixmap] = None,
        icon_error: Optional[QPixmap] = None,
        hide_if_all_empty=False,
    ):
        super().__init__()
        self.line_edits = line_edits
        self.hide_if_all_empty = hide_if_all_empty

        # icons
        style = self.style() or QStyle()
        self.icons = {
            AnalyzerState.Valid: (
                icon_OK
                if icon_OK
                else QPixmap(style.standardPixmap(QStyle.StandardPixmap.SP_DialogApplyButton))
            ),
            AnalyzerState.Warning: (
                icon_warning
                if icon_warning
                else QPixmap(style.standardPixmap(QStyle.StandardPixmap.SP_MessageBoxWarning))
            ),
            AnalyzerState.Invalid: (
                icon_error
                if icon_error
                else QPixmap(style.standardPixmap(QStyle.StandardPixmap.SP_MessageBoxCritical))
            ),
        }

        # Setup layout
        layout: QHBoxLayout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)  # Left, Top, Right, Bottom margins

        # SVG label for icons
        self.icon_label: QLabel = QLabel()
        self.icon_label.setAlignment(Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignRight)

        # title label
        self.title_label: QLabel = QLabel()
        self.title_label.setAlignment(Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignLeft)

        # Text label
        self.text_label: ElidedLabel = ElidedLabel()
        self.text_label.setAlignment(Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignLeft)

        # Add widgets to layout
        layout.addWidget(self.icon_label)
        layout.addItem(QSpacerItem(10, 1, QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed))
        layout.addWidget(self.title_label)
        layout.addWidget(self.text_label)

        # Connect changes in line edits
        for line_edit in self.line_edits:
            line_edit.textChanged.connect(self.updateUi)

        # Initial update
        self.updateUi()

    def updateUi(self):
        self.update_status()
        self.update_label_text()

    def get_analysis_list(self, min_state=AnalyzerState.Valid) -> List[AnalyzerMessage]:
        analysis_list = []
        for le in self.line_edits:
            analyzer = le.analyzer()
            if not analyzer:
                continue
            analysis = analyzer.analyze(le.text())
            if analysis.state >= min_state:
                analysis_list.append(analysis)
        return analysis_list

    def get_worst_analysis(self) -> AnalyzerMessage:
        return BaseAnalyzer.worst_message(self.get_analysis_list())

    def update_status(self) -> None:
        """Update icon based on line edits' contents and validation."""
        self.icon_label.setPixmap(self.icons[self.get_worst_analysis().state])
        self.icon_label.setToolTip(
            "\n".join([str(analysis) for analysis in self.get_analysis_list(min_state=AnalyzerState.Warning)])
        )

        if self.hide_if_all_empty:
            self.setHidden(all([le.text() == "" for le in self.line_edits]))

    def update_label_text(self) -> None:
        """Update text label to show text of all line edits formatted with their object names."""
        titles = [f"{le.objectName()}:" for le in self.line_edits]
        self.title_label.setText("\n".join(titles))

        texts: List[str] = [le.text() for le in self.line_edits]
        self.text_label.setText("\n".join(texts))
        self.text_label.setToolTip("\n".join(texts))


if __name__ == "__main__":

    class CustomIntAnalyzer(BaseAnalyzer):
        """Custom validator that allows any input but validates numeric input."""

        def analyze(self, input: str, pos: int = 0) -> AnalyzerMessage:
            if input.isdigit():
                return AnalyzerMessage("ok", AnalyzerState.Valid)
            elif not input:
                return AnalyzerMessage("empty", AnalyzerState.Warning)
            return AnalyzerMessage("invalid", AnalyzerState.Invalid)

    def setup_line_edit(line_edit: Union[AnalyzerLineEdit, AnalyzerTextEdit]):
        """Set up a QLineEdit with a custom validator that allows all inputs and styles the QLineEdit based on validity."""
        analyzer = CustomIntAnalyzer()
        line_edit.setAnalyzer(analyzer)
        # line_edit.textChanged.connect(lambda text, le=line_edit, val=validator: validate_input(le, val))

    def validate_input(line_edit: Union[AnalyzerLineEdit, AnalyzerTextEdit], analyzer: CustomIntAnalyzer):
        """Update the line edit style based on validation."""
        analysis = analyzer.analyze(line_edit.text(), 0)
        if analysis.state == AnalyzerState.Warning:
            line_edit.setStyleSheet(f"{line_edit.__class__.__name__}" + " { background-color: #ff6c54; }")
        else:
            line_edit.setStyleSheet(f"")

    app = QApplication([])
    le1 = AnalyzerLineEdit()
    le1.setObjectName("Field 1")
    setup_line_edit(le1)
    le2 = AnalyzerLineEdit()
    le2.setObjectName("Field 2")
    setup_line_edit(le2)
    le3 = AnalyzerLineEdit()
    le3.setObjectName("Field 3")
    setup_line_edit(le3)

    window = AnalyzerIndicator([le1, le2, le3])
    form = QFormLayout()
    form.addRow(le1.objectName() + ":", le1)
    form.addRow(le2.objectName() + ":", le2)
    form.addRow(le3.objectName() + ":", le3)
    form.addRow(window)

    main_widget = QWidget()
    main_widget.setLayout(form)
    main_widget.show()

    app.exec()



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from abc import abstractmethod
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Union

logger = logging.getLogger(__name__)


import copy
import enum
import json
import os

# from https://stackoverflow.com/questions/2490334/simple-way-to-encode-a-string-according-to-a-password
import secrets
from base64 import urlsafe_b64decode as b64d
from base64 import urlsafe_b64encode as b64e
from typing import Callable, Dict, Iterable, Type

import bdkpython as bdk
from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from packaging import version


def varnames(method: Callable) -> Iterable[str]:
    return method.__code__.co_varnames[: method.__code__.co_argcount]


def filtered_dict(d: Dict, allowed_keys: Iterable[str]) -> Dict:
    return {k: v for k, v in d.items() if k in allowed_keys}


def filtered_for_init(d: Dict, cls: Type) -> Dict:
    return filtered_dict(d, varnames(cls.__init__))


class Encrypt:
    def _derive_key(self, password: bytes, salt: bytes, iterations: int) -> bytes:
        """Derive a secret key from a given password and salt."""
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA512(),
            length=32,
            salt=salt,
            iterations=iterations,
            backend=default_backend(),
        )
        return b64e(kdf.derive(password))

    def password_encrypt(self, message: bytes, password: str, iterations: int = 100_000) -> bytes:
        salt = secrets.token_bytes(16)
        key = self._derive_key(password.encode(), salt, iterations)
        return b64e(
            b"%b%b%b"
            % (
                salt,
                iterations.to_bytes(4, "big"),
                b64d(Fernet(key).encrypt(message)),
            )
        )

    def password_decrypt(self, token: bytes, password: str) -> bytes:
        decoded = b64d(token)
        salt, iter, token = decoded[:16], decoded[16:20], b64e(decoded[20:])
        iterations = int.from_bytes(iter, "big")
        if iterations > 1e6:
            raise Exception("Error in decrypting")
        key = self._derive_key(password.encode(), salt, iterations)
        return Fernet(key).decrypt(token)


class Storage:
    def __init__(self) -> None:
        self.encrypt = Encrypt()

    def save(self, message: str, filename: str, password: Optional[str] = None) -> None:
        token = self.encrypt.password_encrypt(message.encode(), password) if password else message.encode()

        with open(filename, "wb") as f:
            f.write(token)

    @classmethod
    def has_password(cls, filename: str) -> bool:
        with open(filename, "rb") as f:
            token = f.read()

        if token.decode()[0] == "{":
            return False

        return True

    def load(self, filename: str, password: Optional[str] = None) -> str:
        with open(filename, "rb") as f:
            token = f.read()

        if not password:
            logger.debug(f"Opening {filename} without password")
            return token.decode()
        else:
            logger.debug(f"Decrypting {filename}")
            return self.encrypt.password_decrypt(token, password).decode()


class ClassSerializer:
    @classmethod
    def general_deserializer(cls, known_classes, class_kwargs) -> Callable:
        def deserializer(dct) -> Dict:
            cls_string = dct.get("__class__")  # e.g. KeyStore
            if cls_string:
                if cls_string in known_classes:
                    obj_cls = known_classes[cls_string]
                    if hasattr(obj_cls, "from_dump"):  # is there KeyStore.from_dump ?
                        if class_kwargs.get(cls_string):  #  apply additional arguments to the class from_dump
                            dct.update(class_kwargs.get(cls_string))
                        return obj_cls.from_dump(
                            dct, class_kwargs=class_kwargs
                        )  # do: KeyStore.from_dump(**dct)
                    else:
                        raise Exception(f"{obj_cls} doesnt have a from_dump classmethod.")
                else:
                    raise Exception(
                        f"""{cls_string} not in known_classes {known_classes}."""
                        """Did you add the following to the child class?
                                            VERSION = "0.0.1"
                                            known_classes = {
                                                **BaseSaveableClass.known_classes,
                                            }"""
                        f"""And did you add
                                       "cls_string":{cls_string}
                                       to the parent BaseSaveableClass ?
    """
                    )
            elif dct.get("__enum__"):
                obj_cls = known_classes.get(dct["name"])
                if obj_cls:
                    return getattr(obj_cls, dct["value"])
                else:
                    raise Exception(f"Could not do from_dump(**{dct})")

            # For normal cases, json.loads will handle default JSON data types
            # No need to use json.Decoder here, just return the dictionary as-is
            return dct

        return deserializer

    @classmethod
    def general_serializer(cls, obj):
        if isinstance(obj, enum.Enum):
            return {"__enum__": True, "name": obj.__class__.__name__, "value": obj.name}
        if isinstance(obj, BaseSaveableClass):
            return obj.dump()
        # Fall back to the default JSON serializer
        return json.JSONEncoder().default(obj)


class BaseSaveableClass:
    known_classes: Dict[str, Any] = {"Network": bdk.Network}
    VERSION = "0.0.0"

    @abstractmethod
    def dump(self) -> Dict:
        "Returns the dict"
        d = {}
        d["__class__"] = self.__class__.__name__
        d["VERSION"] = self.VERSION
        return d

    @classmethod
    @abstractmethod
    def from_dump_migration(cls, dct):
        "this class should be overwritten in child classes"
        return dct

    @classmethod
    def _from_dump(cls, dct, class_kwargs=None):
        assert dct.get("__class__") == cls.__name__
        del dct["__class__"]

        if version.parse(cls.VERSION) > version.parse(str(dct.get("VERSION", 0))):
            dct = cls.from_dump_migration(dct)

        if "VERSION" in dct:
            del dct["VERSION"]

    @classmethod
    @abstractmethod
    def from_dump(cls, dct, class_kwargs=None):
        raise NotImplementedError()

    def clone(self, class_kwargs=None):
        return self.from_dump(self.dump(), class_kwargs=class_kwargs)

    def save(self, filename: Union[Path, str], password: Optional[str] = None):
        "Saves the json dumps to a file"
        directory = os.path.dirname(str(filename))
        # Create the directories
        if directory:
            os.makedirs(directory, exist_ok=True)

        storage = Storage()
        storage.save(
            self.dumps(indent=None if password else 4),
            str(filename),
            password=password,
        )

    def __str__(self) -> str:
        return self.dumps()

    def dumps(self, indent=None) -> str:
        "Returns the json representation (recursively)"
        return json.dumps(
            self,
            default=ClassSerializer.general_serializer,
            indent=indent,
            sort_keys=True,
        )

    @staticmethod
    def _flatten_known_classes(known_classes: Dict[str, Any]) -> Dict[str, Any]:
        "Recursively extends the dict to includes all known_classes of known_classes"
        known_classes = known_classes.copy()
        for known_class in list(known_classes.values()):
            if issubclass(known_class, BaseSaveableClass):
                known_classes.update(BaseSaveableClass._flatten_known_classes(known_class.known_classes))
        return known_classes

    @classmethod
    def get_known_classes(cls) -> Dict[str, Any]:
        "Gets a flattened list of known classes that a json deserializer needs to interpet all objects"
        return BaseSaveableClass._flatten_known_classes({cls.__name__: cls})

    @classmethod
    def _from_file(cls, filename: str, password: Optional[str] = None, class_kwargs=None):
        """Loads the class from a file. This offers the option of add class_kwargs args

        Args:
            filename (str): _description_
            password (Optional[str], optional): _description_. Defaults to None.
            class_kwargs (_type_, optional):  example:  class_kwargs= {'Wallet':{'config':config}}. Defaults to None.

        Returns:
            _type_: _description_
        """
        class_kwargs = class_kwargs if class_kwargs else {}
        storage = Storage()

        json_string = storage.load(filename, password=password)

        instance = json.loads(
            json_string,
            object_hook=ClassSerializer.general_deserializer(cls.get_known_classes(), class_kwargs),
        )
        return instance


class SaveAllClass(BaseSaveableClass):
    def dump(self):
        d = super().dump()
        d.update(copy.deepcopy(self.__dict__))
        return d

    @classmethod
    def from_dump(cls, dct: Dict, class_kwargs=None):
        super()._from_dump(dct, class_kwargs=class_kwargs)
        return cls(**filtered_for_init(dct, cls))



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Tuple

from bitcoin_qr_tools.qr_widgets import EnlargableImageWidget

from bitcoin_safe.gui.qt.qr_types import QrType, QrTypes
from bitcoin_safe.gui.qt.synced_tab_widget import SyncedTabWidget
from bitcoin_safe.pdfrecovery import TEXT_24_WORDS

logger = logging.getLogger(__name__)
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import QLabel, QVBoxLayout, QWidget

from .util import screenshot_path


@dataclass
class HardwareSigner:
    name: str
    display_name: str
    usb_preferred: bool
    qr_type: Optional[QrType] = None

    @property
    def generate_seed_png(self):
        return f"{self.name}-generate-seed.png"

    @property
    def wallet_export_png(self):
        return f"{self.name}-wallet-export.png"

    @property
    def view_seed_png(self):
        return f"{self.name}-view-seed.png"

    @property
    def register_multisig_decriptor_png(self):
        return f"{self.name}-register-multisig-decriptor.png"


class HardwareSigners:
    coldcard = HardwareSigner("coldcard", "Coldcard - Mk4", usb_preferred=False)
    q = HardwareSigner("q", "Coldcard - Q", usb_preferred=False, qr_type=QrTypes.bbqr)
    bitbox02 = HardwareSigner("bitbox02", "Bitbox02", usb_preferred=True)
    specterdiy = HardwareSigner(
        "specterdiy", "Specter DIY", usb_preferred=False, qr_type=QrTypes.specterdiy_descriptor_export
    )
    jade = HardwareSigner("jade", "Jade", usb_preferred=True)
    foundation_passport = HardwareSigner(
        "passport", "Foundation - Passport", usb_preferred=False, qr_type=QrTypes.ur
    )


class ScreenshotsTutorial(QWidget):
    enabled_hardware_signers = [
        HardwareSigners.q,
        HardwareSigners.coldcard,
        HardwareSigners.bitbox02,
        HardwareSigners.jade,
        HardwareSigners.specterdiy,
    ]

    def __init__(
        self,
        group: str = "tutorial",
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(parent)

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)  # Left, Top, Right, Bottom margins

        self.title = QLabel()
        font = QFont()
        font.setPointSize(12)
        self.title.setFont(font)
        self.title.setWordWrap(True)

        self._layout.addWidget(self.title)
        self.sync_tab = SyncedTabWidget(group=group, parent=self)
        self._layout.addWidget(self.sync_tab)

    def add_image_tab(
        self, image_path: str, tab_title: str, size_hint: Tuple[int, int]
    ) -> Optional[Tuple[EnlargableImageWidget, QWidget]]:
        if not Path(screenshot_path(image_path)).exists():
            logger.warning(
                f"{self.__class__.__name__}:  {screenshot_path(image_path)} doesnt exist. Cannot load the image tab"
            )
            return None
        tab = QWidget()
        tab_layout = QVBoxLayout(tab)
        image_widget = EnlargableImageWidget(size_hint=size_hint)
        image_widget.load_from_file(screenshot_path(image_path))
        tab_layout.addWidget(image_widget)
        self.sync_tab.addTab(tab, tab_title)
        return image_widget, tab

    def set_title(self, text: str) -> None:
        self.title.setText(text)


class ScreenshotsGenerateSeed(ScreenshotsTutorial):
    def __init__(self, group: str = "tutorial", parent: QWidget | None = None) -> None:
        super().__init__(group, parent)

        self.image_widgets: Dict[str, EnlargableImageWidget] = {}
        self.tabs: Dict[str, QWidget] = {}

        for hardware_signer in self.enabled_hardware_signers:
            result = self.add_image_tab(
                hardware_signer.generate_seed_png, hardware_signer.display_name, size_hint=(400, 50)
            )
            if result:
                image_widget, tab = result
                self.image_widgets[hardware_signer.name] = image_widget
                self.tabs[hardware_signer.name] = tab
        self.updateUi()

    def updateUi(self) -> None:
        self.set_title(
            self.tr(
                "Generate {number} secret seed words on each hardware signer and write them on the recovery sheet"
            ).format(number=TEXT_24_WORDS)
        )


class ScreenshotsExportXpub(ScreenshotsTutorial):
    def __init__(self, group: str = "tutorial", parent: QWidget | None = None) -> None:
        super().__init__(group, parent)

        for hardware_signer in self.enabled_hardware_signers:
            self.add_image_tab(
                hardware_signer.wallet_export_png, hardware_signer.display_name, size_hint=(400, 50)
            )
        self.sync_tab.setMinimumSize(800, 500)
        self.updateUi()

    def updateUi(self) -> None:
        self.set_title(self.tr("How-to export the wallet information from the hardware signer"))


class ScreenshotsViewSeed(ScreenshotsTutorial):
    def __init__(
        self,
        title_text=None,
        group: str = "tutorial",
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(group, parent)

        for hardware_signer in self.enabled_hardware_signers:
            self.add_image_tab(
                hardware_signer.view_seed_png, hardware_signer.display_name, size_hint=(400, 50)
            )

        self.title.setWordWrap(True)
        self.updateUi()

    def updateUi(self) -> None:
        self.set_title(
            self.tr(
                "Compare the {number} words on the backup paper to the hardware signer.\nIf you make a mistake here, your money is lost!"
            ).format(number=TEXT_24_WORDS)
        )


class ScreenshotsRegisterMultisig(ScreenshotsTutorial):
    def __init__(
        self,
        group: str = "tutorial",
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(
            group,
            parent,
        )
        self.setMinimumSize(500, 300)

        for hardware_signer in self.enabled_hardware_signers:
            self.add_image_tab(
                hardware_signer.register_multisig_decriptor_png,
                hardware_signer.display_name,
                size_hint=(400, 50),
            )
        self.updateUi()

    def updateUi(self) -> None:
        self.set_title(self.tr("Import the multisig information in the hardware signer"))



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Type

from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtGui import QResizeEvent
from PyQt6.QtWidgets import (
    QApplication,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QSpacerItem,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from bitcoin_safe.gui.qt.data_tab_widget import DataTabWidget, T
from bitcoin_safe.gui.qt.util import read_QIcon


class ExtendedTabWidget(DataTabWidget):
    signal_tab_bar_visibility = pyqtSignal(bool)

    def __init__(self, data_class: Type[T], parent=None) -> None:
        super().__init__(data_class, parent=parent)
        self.set_top_right_widget()

        self.tabBar().installEventFilter(self)  # type: ignore

        self.tabCloseRequested.connect(self.updateLineEditPosition)
        self.currentChanged.connect(self.updateLineEditPosition)

    def eventFilter(self, obj, event) -> bool:
        if obj == self.tabBar() and event.type() == event.Type.Show:
            if self.top_right_widget:
                self.signal_tab_bar_visibility.emit(True)
        elif obj == self.tabBar() and event.type() == event.Type.Hide:
            if self.top_right_widget:
                self.signal_tab_bar_visibility.emit(False)
        return super().eventFilter(obj, event)

    def set_top_right_widget(self, top_right_widget: QWidget | None = None, target_width=150) -> None:
        self.top_right_widget = top_right_widget
        self.target_width = target_width

        # Adjust the size and position of the QLineEdit
        if self.top_right_widget:
            self.top_right_widget.setFixedWidth(self.target_width)
            self.top_right_widget.setParent(self)
            self.updateLineEditPosition()

    def tabInserted(self, index: int) -> None:
        super().tabInserted(index)
        self.updateLineEditPosition()

    def updateLineEditPosition(self) -> None:
        tabBarRect = self.tabBar().geometry()  # type: ignore[union-attr]
        availableWidth = self.width()

        line_width = availableWidth // 2 if availableWidth < 2 * self.target_width else self.target_width

        self.tabBar().setMaximumWidth(availableWidth - line_width - 3)  # type: ignore[union-attr]

        # Update QLineEdit geometry
        lineEditX = self.width() - line_width - 2
        if self.top_right_widget:
            v_margin = (tabBarRect.height() - self.top_right_widget.height()) // 2
            self.top_right_widget.setGeometry(
                lineEditX, tabBarRect.y(), line_width, tabBarRect.height() - v_margin
            )
            self.top_right_widget.setFixedWidth(line_width)  # Ensure fixed width is maintained

    def resizeEvent(self, event: QResizeEvent | None) -> None:
        self.updateLineEditPosition()
        super().resizeEvent(event)


class LoadingWalletTab(QWidget):
    def __init__(self, tabs: DataTabWidget, name: str, focus=True) -> None:
        super().__init__(tabs)
        self.tabs = tabs
        self.name = name
        self.focus = focus

        # Create a QWidget to serve as a container for the QLabel
        self._layout = QVBoxLayout(self)  # Setting the layout directly

        # Create and configure QLabel
        self.emptyLabel = QLabel(self.tr("Loading, please wait..."), self)
        self.emptyLabel.setAlignment(Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter)
        self.emptyLabel.setStyleSheet("font-size: 16pt;")  # Adjust the font size as needed

        # Use spacers to push the QLabel to the center
        spacerTop = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        spacerBottom = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        # Add spacers and label to the layout
        self._layout.addItem(spacerTop)
        self._layout.addWidget(self.emptyLabel)
        self._layout.addItem(spacerBottom)

    def __enter__(self) -> None:
        self.tabs.add_tab(
            tab=self,
            icon=read_QIcon("status_waiting.svg"),
            description=self.name,
            data=None,
            focus=self.focus,
        )
        QApplication.processEvents()

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        idx = self.tabs.indexOf(self)
        if idx is None or idx < 0:
            return
        self.tabs.removeTab(idx)


# Usage example
if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    edit = QLineEdit(f"Ciiiiii")
    tabWidget = ExtendedTabWidget(object)

    # Add tabs with larger widgets
    for i in range(3):
        widget = QWidget()
        layout = QVBoxLayout()
        label = QLabel(f"Content for Tab {i+1}")
        textEdit = QTextEdit(f"This is a larger widget in Tab {i+1}.")
        layout.addWidget(label)
        layout.addWidget(textEdit)
        widget.setLayout(layout)
        tabWidget.addTab(widget, description=f"Tab {i+1}")

    tabWidget.show()
    sys.exit(app.exec())



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import glob
import hashlib
import logging
import os
import platform
import shutil
import subprocess
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pgpy  # Python-native OpenPGP library
import requests

gnupg = None
logger = logging.getLogger(__name__)


@dataclass
class SimpleGPGKey:
    key: str
    repository: str  # org/repo_name
    prefix: str
    manifest_ending: Optional[str] = None

    @staticmethod
    def extract_prefix_and_version(filename: str) -> tuple[Optional[str], Optional[str]]:
        import re

        if filename.endswith(".deb"):
            match = re.search(r"(.+)_(.+?)(?:-.+)?_.*\.deb", filename)
            if match:
                return (match.group(1), match.group(2))

        # try with - separator
        match = re.search(r"(.*?)-([\d\.]+[a-zA-Z0-9]*)", filename)
        if match:
            return (match.group(1), match.group(2))

        # try with _ separator
        match = re.search(r"(.*?)_([\d\.]+[a-zA-Z0-9]*)", filename)
        if match:
            return (match.group(1), match.group(2))
        return (None, None)

    def get_tag_if_mine(self, filename: str) -> Optional[str]:
        prefix, version = self.extract_prefix_and_version(filename)
        if prefix and prefix.lower() == self.prefix.lower():
            return version
        return None


@dataclass
class Asset:
    tag: str
    url: str
    name: str


class GitHubAssetDownloader:
    def __init__(self, repository: str) -> None:
        self.repository = repository
        logger.debug(f"initialized {self}")

    def _get_assets(self, api_url) -> List[Asset]:
        try:
            logger.debug(f"Get assets from {api_url}")
            response = requests.get(api_url, timeout=2)
            response.raise_for_status()
            assets = response.json().get("assets", [])

            return [
                Asset(
                    tag=response.json().get("tag_name", ""),
                    url=asset["browser_download_url"],
                    name=asset["name"],
                )
                for asset in assets
            ]
        except requests.RequestException as e:
            logger.error(f"Failed to download: {api_url}")
            return []

    def get_assets_by_tag(self, tag: str) -> List[Asset]:
        return self._get_assets(f"https://api.github.com/repos/{self.repository}/releases/tags/{tag}")

    def get_assets_latest(self) -> List[Asset]:
        return self._get_assets(f"https://api.github.com/repos/{self.repository}/releases/latest")


class SignatureVerifyer:
    def __init__(
        self,
        list_of_known_keys: Optional[List[SimpleGPGKey]],
    ) -> None:
        self.list_of_known_keys = list_of_known_keys if list_of_known_keys else []
        self.public_keys: Dict[str, pgpy.PGPKey] = {}
        self.import_known_keys()

    def import_public_key_file(self, path: Path) -> pgpy.PGPKey:
        with open(str(path), "r") as file:
            return self.import_public_key_block(file.read())

    def import_public_key_block(self, public_key_block: str) -> pgpy.PGPKey:
        """
        Import a public key block and return its fingerprint.

        :param public_key_block: The public key block in ASCII armor format.
        :return: The fingerprint of the imported public key.
        """

        result = pgpy.PGPKey.from_blob(public_key_block)
        public_key = None
        if isinstance(result, pgpy.PGPKey):
            public_key = result
        elif isinstance(result, tuple):
            public_key, _ = result

        if isinstance(public_key, pgpy.PGPKey):
            fingerprint = str(public_key.fingerprint)
            self.public_keys[fingerprint] = public_key
            logger.info(f"Public key imported with fingerprint: {public_key.fingerprint}")
            return public_key

        raise Exception(f"Could not process result f{result}")

    def import_known_keys(self) -> None:
        for key in self.list_of_known_keys:
            self.import_public_key_block(key.key)

    @staticmethod
    def verify_manifest_hashes(manifest_file: Path) -> bool:
        # Check if the manifest file exists
        if not manifest_file.exists():
            print("Manifest file does not exist.")
            return False

        file_found = False  # To track if at least one file is found

        with manifest_file.open("r") as file:
            for line in file:
                # Extract expected hash and file name from each line
                parts = line.strip().split(" *")
                if len(parts) != 2:
                    print("Invalid line format in manifest:", line)
                    continue  # Skip invalid lines

                expected_hash, file_name = parts
                file_path = manifest_file.parent / file_name

                # Check if the file exists
                if not file_path.exists():
                    logger.debug(f"File {file_name} listed in the manifest does not exist.")
                    continue  # It's okay if some files don't exist, but at least one must be found

                file_found = True  # Mark that we've found at least one file

                # Compute the file's SHA-256 hash
                file_hash = hashlib.sha256()
                with open(file_path, "rb") as f:
                    for chunk in iter(lambda: f.read(4096), b""):
                        file_hash.update(chunk)

                # Compare the computed hash with the expected hash
                if file_hash.hexdigest() != expected_hash:
                    logger.error(f"Hash mismatch for {file_name}.")
                    return False

        if not file_found:
            print("None of the files listed in the manifest are present.")
            return False  # Return False if none of the files are found

        return True

    def get_valid_manifest(self, signature_file: Path) -> Optional[Path]:
        manifest = signature_file.parent / signature_file.name.removesuffix(".asc")
        return manifest if self.verify_manifest_hashes(manifest) else None

    def verify_signature(self, binary_filename: Path, expected_public_key: SimpleGPGKey) -> bool:
        signature_file = self.get_signature_file(binary_filename)
        if not signature_file:
            signature_file = self.get_signature_from_web(binary_filename)
            if not signature_file:
                logger.error("Could not download signature file")
                return False

        key = self.get_key(binary_filename)
        public_key = self.import_public_key_block(expected_public_key.key)
        if not public_key:
            logger.error("Expected public key not found.")
            return False

        binary_file_to_check = binary_filename
        if key and key.manifest_ending:
            manifest_file = self.get_valid_manifest(signature_file)
            if not manifest_file:
                return False
            binary_file_to_check = manifest_file

        verification_result = self._verify_file(
            public_key=public_key, binary_file=binary_file_to_check, signature_file=signature_file
        )

        return verification_result

    def _verify_file(
        self, public_key: pgpy.PGPKey, binary_file: Union[Path, str], signature_file: Union[Path, str]
    ) -> bool:
        try:
            with open(str(signature_file), "rb") as sig_file:
                signature = pgpy.PGPSignature.from_blob(sig_file.read())

            with open(str(binary_file), "rb") as bin_file:
                message_data = bin_file.read()

            pgpmessage = pgpy.PGPMessage.new(message_data, file=True)

            # Verify the signature
            verify_result = public_key.verify(pgpmessage.message, signature)
            if not verify_result:
                return False
            good_signatures = list(verify_result.good_signatures)
            bad_signatures = list(verify_result.bad_signatures)
            # 1 single bad signature creates a False result
            return bool(good_signatures) and not bool(bad_signatures)
        except Exception as e:
            logger.error(f"Verification failed: {e}")
            return False

    @staticmethod
    def _get_asset_url(assets: List[Asset], ending: str) -> Optional[str]:
        for asset in assets:
            if asset.name.endswith(ending):
                return asset.url
        return None

    @staticmethod
    def _download_file(download_url: str, filename: Path) -> Path:
        sig_response = requests.get(download_url, timeout=2)
        sig_response.raise_for_status()
        with open(filename, "wb") as f:
            f.write(sig_response.content)
        return filename

    @staticmethod
    def _download_asset_file(
        assets: List[Asset], target_directory: Path, asset_ending: str
    ) -> Optional[Path]:
        if url := SignatureVerifyer._get_asset_url(assets, asset_ending):
            url_filename = Path(url).name
            filename = target_directory / url_filename
            SignatureVerifyer._download_file(url, filename)
            return filename
        return None

    def get_signature_file(self, binary_filename: Path) -> Optional[Path]:
        key = self.get_key(binary_filename)
        if key and key.manifest_ending:
            manifest_files = glob.glob(str(binary_filename.parent / f"*{key.manifest_ending}"))
            sig_files = glob.glob(str(binary_filename.parent / f"*{key.manifest_ending}.asc"))

            if not manifest_files or not sig_files:
                return None
            return Path(sig_files[0])
        else:
            return binary_filename.parent / (binary_filename.name + ".asc")

    def is_signature_file_available(self, binary_filename: Path) -> bool:
        sig_file = self.get_signature_file(binary_filename)
        return sig_file.exists() if sig_file else False

    def get_key(self, binary_filename: Path) -> Optional[SimpleGPGKey]:
        for key in self.list_of_known_keys:
            tag = key.get_tag_if_mine(binary_filename.name)
            if tag:
                return key
        return None

    def get_signature_from_web(self, binary_filename: Path) -> Optional[Path]:
        sig_filename = None
        key = self.get_key(binary_filename)
        if not key:
            return None
        tag = key.get_tag_if_mine(binary_filename.name)
        if not tag:
            return None

        assets = GitHubAssetDownloader(repository=key.repository).get_assets_by_tag(tag)
        if assets:
            if key.manifest_ending:
                self._download_asset_file(
                    assets, target_directory=binary_filename.parent, asset_ending=key.manifest_ending
                )
                sig_filename = self._download_asset_file(
                    assets,
                    target_directory=binary_filename.parent,
                    asset_ending=f"{key.manifest_ending}.asc",
                )
            else:
                sig_filename = self._download_asset_file(
                    assets,
                    target_directory=binary_filename.parent,
                    asset_ending=f"{binary_filename.name}.asc",
                )
        return sig_filename

    def get_key_meta_info(self, key: SimpleGPGKey) -> Dict:
        """
        Retrieve meta-information about a key given its fingerprint.

        :param key: The GPGKey object containing the key data.
        :return: A dictionary containing the key's meta-information.
        """
        public_key = self.import_public_key_block(key.key)

        key_info = {
            "fingerprint": public_key.fingerprint,
            "userids": [uid.name for uid in public_key.userids],
            "created": public_key.created,
            "algorithm": public_key.key_algorithm.name,
        }
        return key_info


class GPGSignatureVerifyer:
    def __init__(
        self,
        list_of_known_keys: Optional[List[SimpleGPGKey]],
    ) -> None:
        self.list_of_known_keys = list_of_known_keys if list_of_known_keys else []
        self._gpg: Any = None
        self.import_known_keys()

    @staticmethod
    def _get_gpg_path() -> Optional[str]:
        command = "gpg"
        gpg_path = shutil.which(command)
        if gpg_path:
            try:
                subprocess.run(
                    [command, "--version"], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                )
                return command
            except subprocess.CalledProcessError:
                return None
        elif platform.system() == "Windows":
            import winreg

            try:
                # Open the registry key
                with winreg.OpenKey(  # type: ignore
                    winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\WOW6432Node\GnuPG", 0, winreg.KEY_READ  # type: ignore
                ) as key:
                    # Read the value of the Install Directory
                    install_dir, _ = winreg.QueryValueEx(key, "Install Directory")  # type: ignore
                    gpg_path = os.path.join(install_dir, "bin", "gpg.exe")
                    if os.path.isfile(gpg_path):
                        return gpg_path
            except FileNotFoundError:
                pass
            except OSError as e:
                print(f"Error accessing registry: {e}")
            return None
        return None

    @classmethod
    def is_gnupg_installed(cls) -> bool:
        return bool(cls._get_gpg_path())

    @property
    def gpg(self) -> Any:
        if self._gpg:
            return self._gpg

        if self.is_gnupg_installed():
            global gnupg
            import gnupg as _gnupg

            gnupg = _gnupg

            gpg_path = self._get_gpg_path()
            if not gpg_path:
                raise EnvironmentError("GnuPG path could not be determined.")
            self._gpg = gnupg.GPG(gpgbinary=gpg_path)
            return self._gpg
        else:
            raise EnvironmentError("GnuPG is not installed on this system.")

    def import_known_keys(self) -> None:
        for key in self.list_of_known_keys:
            self.gpg.import_keys(key.key)

    def import_public_key_block(self, public_key_block: str) -> str:
        """
        Import a public key block into the GnuPG keyring and return its fingerprint.

        :param public_key_block: The public key block in ASCII armor format.
        :return: The fingerprint of the imported public key.
        """
        import_result = self.gpg.import_keys(public_key_block)

        if not import_result.results:
            raise ValueError("Failed to import public key.")

        # The 'fingerprint' attribute is available in the import result.
        # For multiple keys, this simple implementation assumes you're interested in the first.
        fingerprint = import_result.results[0]["fingerprint"]
        logger.info(f"Public key imported with fingerprint: {fingerprint}")
        return fingerprint


class SignatureSigner(GPGSignatureVerifyer):
    def __init__(
        self,
        version: str,
        app_name: str,
        list_of_known_keys: Optional[List[SimpleGPGKey]],
        file_dir: str = "dist/",
    ) -> None:
        super().__init__(list_of_known_keys=list_of_known_keys)

        self.file_dir = file_dir
        self.version = version
        self.app_name = app_name

    def get_files_to_sign(self) -> List[Path]:
        return [
            f
            for f in Path(self.file_dir).iterdir()
            if f.is_file() and (not f.name.endswith(".asc")) and (not f.name.startswith("."))
        ]

    def sign_files(self, key: SimpleGPGKey) -> List[Path]:
        my_fingerprint = self.import_public_key_block(key.key)
        files = self.get_files_to_sign()
        for file_path in files:

            self.gpg.sign_file(
                open(file_path, "rb"), keyid=my_fingerprint, detach=True, output=str(file_path) + ".asc"
            )
            logger.info(f"File signed: {file_path.name}.asc")
        return files


@dataclass
class KnownGPGKeys:

    andreasgriffin = SimpleGPGKey(
        key="""
-----BEGIN PGP PUBLIC KEY BLOCK-----
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=aRyX
-----END PGP PUBLIC KEY BLOCK-----               
                   """,
        repository="andreasgriffin/bitcoin-safe",
        prefix="bitcoin-safe",
    )
    craigraw = SimpleGPGKey(
        repository="sparrowwallet/sparrow",
        prefix="sparrow",
        manifest_ending="manifest.txt",
        key="""
-----BEGIN PGP PUBLIC KEY BLOCK-----
Comment: GPGTools - http://gpgtools.org


EmR2yA==
=t5JY
-----END PGP PUBLIC KEY BLOCK-----

               """,
    )

    @classmethod
    def all(cls) -> List[SimpleGPGKey]:
        return [v for v in cls.__dict__.values() if isinstance(v, SimpleGPGKey)]



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# Original Version from:
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2015 Thomas Voegtlin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import os
import tempfile

from PyQt6.QtGui import QFont, QFontMetrics

from bitcoin_safe.config import UserConfig
from bitcoin_safe.gui.qt.wrappers import Menu
from bitcoin_safe.mempool import MempoolData
from bitcoin_safe.psbt_util import FeeInfo
from bitcoin_safe.pythonbdk_types import Balance, Recipient

logger = logging.getLogger(__name__)

import datetime
import enum
from enum import IntEnum
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple

import bdkpython as bdk
from bitcoin_qr_tools.data import Data
from PyQt6.QtCore import QMimeData, QModelIndex, QPoint, QSize, Qt, pyqtSignal
from PyQt6.QtGui import (
    QBrush,
    QColor,
    QDragEnterEvent,
    QDragMoveEvent,
    QDropEvent,
    QFont,
    QStandardItem,
)
from PyQt6.QtWidgets import QAbstractItemView, QFileDialog, QPushButton, QStyle, QWidget

from ...i18n import translate
from ...signals import Signals, UpdateFilter, UpdateFilterReason
from ...util import Satoshis, block_explorer_URL, confirmation_wait_formatted
from ...wallet import (
    ToolsTxUiInfo,
    TxConfirmationStatus,
    TxStatus,
    Wallet,
    get_wallet,
    get_wallets,
)
from .category_list import CategoryEditor
from .dialog_import import file_to_str
from .my_treeview import (
    MyItemDataRole,
    MySortModel,
    MyStandardItemModel,
    MyTreeView,
    TreeViewWithToolbar,
    needs_frequent_flag,
)
from .taglist import AddressDragInfo
from .util import Message, MessageType, read_QIcon, sort_id_to_icon, webopen


class AddressUsageStateFilter(IntEnum):
    ALL = 0
    UNUSED = 1
    FUNDED = 2
    USED_AND_EMPTY = 3
    FUNDED_OR_UNUSED = 4

    def ui_text(self) -> str:
        return {
            self.ALL: translate("hist_list", "All status"),
            self.UNUSED: translate("hist_list", "Unused"),
            self.FUNDED: translate("hist_list", "Funded"),
            self.USED_AND_EMPTY: translate("hist_list", "Used"),
            self.FUNDED_OR_UNUSED: translate("hist_list", "Funded or Unused"),
        }[self]


class AddressTypeFilter(IntEnum):
    ALL = 0
    RECEIVING = 1
    CHANGE = 2

    def ui_text(self) -> str:
        return {
            self.ALL: translate("hist_list", "All types"),
            self.RECEIVING: translate("hist_list", "Receiving"),
            self.CHANGE: translate("hist_list", "Change"),
        }[self]


class HistList(MyTreeView):
    signal_tag_dropped = pyqtSignal(AddressDragInfo)

    show_change: AddressTypeFilter
    show_used: AddressUsageStateFilter

    class Columns(MyTreeView.BaseColumnsEnum):
        TXID = enum.auto()
        WALLET_ID = enum.auto()
        STATUS = enum.auto()
        CATEGORIES = enum.auto()
        LABEL = enum.auto()
        AMOUNT = enum.auto()
        BALANCE = enum.auto()

    filter_columns = [
        Columns.WALLET_ID,
        Columns.STATUS,
        Columns.CATEGORIES,
        Columns.LABEL,
        Columns.AMOUNT,
        Columns.TXID,
    ]

    column_alignments = {
        Columns.WALLET_ID: Qt.AlignmentFlag.AlignCenter,
        Columns.STATUS: Qt.AlignmentFlag.AlignCenter,
        Columns.CATEGORIES: Qt.AlignmentFlag.AlignCenter,
        Columns.LABEL: Qt.AlignmentFlag.AlignVCenter,
        Columns.AMOUNT: Qt.AlignmentFlag.AlignRight,
        Columns.BALANCE: Qt.AlignmentFlag.AlignRight,
    }

    def __init__(
        self,
        fx,
        config,
        signals: Signals,
        mempool_data: MempoolData,
        wallet_id: str,
        hidden_columns: List[int] | None = None,
        column_widths: Optional[Dict[MyTreeView.BaseColumnsEnum, int]] = None,
        address_domain: List[str] | None = None,
    ) -> None:
        super().__init__(
            config=config,
            stretch_column=HistList.Columns.LABEL,
            editable_columns=[HistList.Columns.LABEL],
            column_widths=column_widths,
            signals=signals,
            sort_column=HistList.Columns.STATUS,
            sort_order=Qt.SortOrder.DescendingOrder,
        )
        self.fx = fx
        self.mempool_data = mempool_data
        self.address_domain = address_domain
        self.hidden_columns = hidden_columns if hidden_columns else []
        self._tx_dict: Dict[str, Tuple[Wallet, bdk.TransactionDetails]] = {}  # txid -> wallet, tx
        self.signals = signals
        self.wallet_id = wallet_id
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setSortingEnabled(True)
        self.show_change = AddressTypeFilter.ALL  # type: AddressTypeFilter
        self.show_used = AddressUsageStateFilter.ALL  # type: AddressUsageStateFilter
        self.balance = 0
        # self.change_button = QComboBox(self)
        # self.change_button.currentIndexChanged.connect(self.toggle_change)
        # for (
        #     addr_type
        # ) in AddressTypeFilter.__members__.values():  # type: AddressTypeFilter
        #     self.change_button.addItem(addr_type.ui_text())
        # self.used_button = QComboBox(self)
        # self.used_button.currentIndexChanged.connect(self.toggle_used)
        # for (
        #     addr_usage_state
        # ) in (
        #     AddressUsageStateFilter.__members__.values()
        # ):  # type: AddressUsageStateFilter
        #     self.used_button.addItem(addr_usage_state.ui_text())
        self._source_model = MyStandardItemModel(
            self,
            drag_key="txids",
            drag_keys_to_file_paths=self.drag_keys_to_file_paths,
        )
        self.proxy = MySortModel(
            self, source_model=self._source_model, sort_role=MyItemDataRole.ROLE_SORT_ORDER
        )
        self.setModel(self.proxy)
        self.update_content()
        self.signals.wallet_signals[self.wallet_id].updated.connect(self.update_with_filter)
        self.signals.language_switch.connect(self.update)

    def get_file_data(self, txid: str) -> Optional[Data]:
        for wallet in get_wallets(self.signals):
            txdetails = wallet.get_tx(txid)
            if txdetails:
                return Data.from_tx(txdetails.transaction)
        return None

    def drag_keys_to_file_paths(
        self, drag_keys: Iterable[str], save_directory: Optional[str] = None
    ) -> List[str]:
        file_urls = []

        # Iterate through indexes to fetch serialized data using drag keys
        for key in drag_keys:
            # Fetch the serialized data using the drag_key
            data = self.get_file_data(key)
            if not data:
                continue

            if save_directory:
                file_path = os.path.join(save_directory, f"{key}.tx")
                file_descriptor = os.open(file_path, os.O_CREAT | os.O_WRONLY)
            else:
                # Create a temporary file
                file_descriptor, file_path = tempfile.mkstemp(
                    suffix=f".tx",
                    prefix=f"{key} ",
                )

            data.write_to_filedescriptor(file_descriptor)

            # Add the file URL to the list
            file_urls.append(file_path)

        return file_urls

    def _acceptable_mime_data(self, mime_data: QMimeData) -> bool:
        if mime_data and self.get_json_mime_data(mime_data) is not None:
            return True
        if mime_data and mime_data.hasUrls():
            return True
        return False

    def dragEnterEvent(self, event: QDragEnterEvent | None) -> None:
        super().dragEnterEvent(event)
        if not event or event.isAccepted():
            return

        mime_data = event.mimeData()
        if mime_data and self._acceptable_mime_data(mime_data):
            event.acceptProposedAction()
        else:
            event.ignore()

    def dragMoveEvent(self, event: QDragMoveEvent | None) -> None:
        super().dragMoveEvent(event)
        if not event or event.isAccepted():
            return

        mime_data = event.mimeData()
        if mime_data and self._acceptable_mime_data(mime_data):
            event.acceptProposedAction()
        else:
            event.ignore()

    def dropEvent(self, event: QDropEvent | None) -> None:
        # handle dropped files
        super().dropEvent(event)
        if not event or event.isAccepted():
            return

        index = self.indexAt(event.position().toPoint())
        if not index.isValid():
            # Handle the case where the drop is not on a valid index
            return

        mime_data = event.mimeData()
        if mime_data:
            json_mime_data = self.get_json_mime_data(mime_data)
            if json_mime_data is not None:
                model = self.model()
                hit_address = model.data(model.index(index.row(), self.key_column))
                if json_mime_data.get("type") == "drag_tag":
                    if hit_address is not None:
                        drag_info = AddressDragInfo([json_mime_data.get("tag")], [hit_address])
                        logger.debug(f"drag_info {drag_info}")
                        self.signal_tag_dropped.emit(drag_info)
                    event.accept()
                    return

            elif mime_data.hasUrls():
                # Iterate through the list of dropped file URLs
                for url in mime_data.urls():
                    # Convert URL to local file path
                    file_path = url.toLocalFile()
                    self.signals.open_tx_like.emit(file_to_str(file_path))

        event.ignore()

    def on_double_click(self, idx: QModelIndex) -> None:
        txid = self.get_role_data_for_current_item(col=self.key_column, role=MyItemDataRole.ROLE_KEY)
        wallet, tx_details = self._tx_dict[txid]
        self.signals.open_tx_like.emit(tx_details)

    def toggle_change(self, state: int) -> None:
        if state == self.show_change:
            return
        self.show_change = AddressTypeFilter(state)
        self.update_content()

    def toggle_used(self, state: int) -> None:
        if state == self.show_used:
            return
        self.show_used = AddressUsageStateFilter(state)
        self.update_content()

    def update_with_filter(self, update_filter: UpdateFilter) -> None:
        if update_filter.refresh_all:
            return self.update_content()
        logger.debug(f"{self.__class__.__name__} update_with_filter {update_filter}")

        def categories_intersect(model: MyStandardItemModel, row) -> Set:
            return set(model.data(model.index(row, self.Columns.CATEGORIES))).intersection(
                set(update_filter.categories)
            )

        def tx_involves_address(txid) -> Set[str]:
            (wallet, tx) = self._tx_dict[txid]
            fulltxdetail = wallet.get_dict_fulltxdetail().get(txid)
            if not fulltxdetail:
                return set()
            return update_filter.addresses.intersection(fulltxdetail.involved_addresses())

        logger.debug(f"{self.__class__.__name__}  update_with_filter {update_filter}")
        self._before_update_content()

        log_info = []
        model = self._source_model
        # Select rows with an ID in id_list
        for row in range(model.rowCount()):
            txid = model.data(model.index(row, self.Columns.TXID))

            if (
                update_filter.reason == UpdateFilterReason.ChainHeightAdvanced
                and model.data(
                    model.index(row, self.key_column), role=MyItemDataRole.ROLE_FREQUENT_UPDATEFLAG
                )
            ) or any(
                [txid in update_filter.txids, categories_intersect(model, row), tx_involves_address(txid)]
            ):
                log_info.append((row, txid))
                self.refresh_row(txid, row)

        logger.debug(f"Updated  {log_info}")

        self._after_update_content()

    def get_headers(self) -> Dict["HistList.Columns", str]:
        return {
            self.Columns.WALLET_ID: self.tr("Wallet"),
            self.Columns.STATUS: self.tr("Status"),
            self.Columns.CATEGORIES: self.tr("Category"),
            self.Columns.LABEL: self.tr("Label"),
            self.Columns.AMOUNT: self.tr("Amount"),
            self.Columns.BALANCE: self.tr("Balance"),
            self.Columns.TXID: self.tr("Txid"),
        }

    def _init_row(
        self, wallet: Wallet, tx: bdk.TransactionDetails, status_sort_index: int, old_balance: int
    ) -> Tuple[List[QStandardItem], int]:
        """

        Returns:
            Tuple[List[QStandardItem] , int]: items, amount
        """

        # WALLET_ID = enum.auto()
        # AMOUNT = enum.auto()
        # BALANCE = enum.auto()
        # TXID = enum.auto()
        self._tx_dict[tx.txid] = (wallet, tx)

        # calculate the amount
        if self.address_domain:
            fulltxdetail = wallet.get_dict_fulltxdetail().get(tx.txid)
            assert fulltxdetail, f"Could not find the transaction for {tx.txid}"
            amount = fulltxdetail.sum_outputs(self.address_domain) - fulltxdetail.sum_inputs(
                self.address_domain
            )
        else:
            amount = int(tx.received - tx.sent)

        new_balance = old_balance + amount

        labels = [""] * len(self.Columns)
        labels[self.Columns.WALLET_ID] = wallet.id
        labels[self.Columns.AMOUNT] = Satoshis(amount, wallet.network).str_as_change()

        labels[self.Columns.BALANCE] = str(Satoshis(new_balance, wallet.network))
        labels[self.Columns.TXID] = tx.txid
        items = [QStandardItem(e) for e in labels]

        items[self.Columns.STATUS].setData(status_sort_index, MyItemDataRole.ROLE_SORT_ORDER)
        items[self.Columns.WALLET_ID].setData(wallet.id, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        items[self.Columns.AMOUNT].setData(amount, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        if amount < 0:
            items[self.Columns.AMOUNT].setData(QBrush(QColor("red")), Qt.ItemDataRole.ForegroundRole)
        items[self.Columns.BALANCE].setData(new_balance, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        items[self.Columns.TXID].setData(tx.txid, MyItemDataRole.ROLE_CLIPBOARD_DATA)

        # align text and set fonts
        # for i, item in enumerate(items):
        #     item.setTextAlignment(Qt.AlignmentFlag.AlignVCenter)
        #     if i in (self.Columns.TXID,):
        #         item.setFont(QFont(MONOSPACE_FONT))

        self.set_editability(items)

        items[self.key_column].setData(tx.txid, MyItemDataRole.ROLE_KEY)

        return items, amount

    def update_content(self) -> None:
        if self.maybe_defer_update():
            return

        self._before_update_content()
        self._tx_dict = {}
        wallets = [
            wallet for wallet in get_wallets(self.signals) if self.wallet_id and wallet.id == self.wallet_id
        ]

        self._source_model.clear()
        self.update_headers(self.get_headers())

        num_shown = 0
        self.balance = 0
        for wallet in wallets:

            txid_domain: Optional[Set[str]] = None
            if self.address_domain:
                txid_domain = set()
                for address in self.address_domain:
                    txid_domain = txid_domain.union(wallet.get_involved_txids(address))

            # always take sorted_delta_list_transactions().new as a start because it is correctly sorted
            for i, tx in enumerate(wallet.sorted_delta_list_transactions()):
                if txid_domain is not None:
                    if tx.txid not in txid_domain:
                        continue

                items, amount = self._init_row(wallet, tx, i, self.balance)
                self.balance += amount

                num_shown += 1
                # add item
                count = self._source_model.rowCount()
                self._source_model.insertRow(count, items)
                self.refresh_row(tx.txid, count)

        super().update_content()
        self._after_update_content()

    def refresh_row(self, key: str, row: int) -> None:
        assert row is not None
        wallet, tx = self._tx_dict[key]
        # STATUS = enum.auto()
        # CATEGORIES = enum.auto()
        # LABEL = enum.auto()

        label = wallet.get_label_for_txid(tx.txid)
        categories = wallet.get_categories_for_txid(tx.txid)
        categories_without_default = set(categories) - set([wallet.labels.get_default_category()])
        category = (
            list(categories_without_default)[0]
            if categories_without_default
            else (categories[0] if categories else "")
        )
        status = TxStatus.from_wallet(tx.txid, wallet)

        fee_info = FeeInfo.from_txdetails(tx)
        estimated_duration_str = confirmation_wait_formatted(
            self.mempool_data.fee_rate_to_projected_block_index(fee_info.fee_rate())
        )
        status_text = (
            datetime.datetime.fromtimestamp(tx.confirmation_time.timestamp).strftime("%Y-%m-%d %H:%M")
            if tx.confirmation_time
            else estimated_duration_str
        )
        status_data = (
            datetime.datetime.fromtimestamp(tx.confirmation_time.height).strftime("%Y-%m-%d %H:%M")
            if tx.confirmation_time
            else TxConfirmationStatus.to_str(status.confirmation_status)
        )
        status_tooltip = (
            self.tr("{number} Confirmations").format(number=status.confirmations())
            if 1 <= status.confirmations() <= 6
            else status_text
        )

        _item = [self._source_model.item(row, col) for col in self.Columns]
        item = [entry for entry in _item if entry]
        if needs_frequent_flag(status=status):
            item[self.key_column].setData(True, role=MyItemDataRole.ROLE_FREQUENT_UPDATEFLAG)
        item[self.Columns.STATUS].setText(status_text)
        item[self.Columns.STATUS].setData(status_data, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        item[self.Columns.STATUS].setIcon(read_QIcon(sort_id_to_icon(status.sort_id())))
        item[self.Columns.STATUS].setToolTip(status_tooltip)
        item[self.Columns.LABEL].setText(label)
        item[self.Columns.LABEL].setData(label, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        item[self.Columns.CATEGORIES].setText(category)
        item[self.Columns.CATEGORIES].setData(categories, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        item[self.Columns.CATEGORIES].setBackground(CategoryEditor.color(category))

    def create_menu(self, position: QPoint) -> Menu:
        menu = Menu()
        # is_multisig = isinstance(self.wallet, Multisig_Wallet)
        selected = self.selected_in_column(self.Columns.TXID)
        if not selected:
            return menu
        multi_select = len(selected) > 1

        _selected_items = [self.item_from_index(item) for item in selected]
        selected_items = [item for item in _selected_items if item]
        txids = [item.text() for item in selected_items if item]
        if not multi_select:
            idx = self.indexAt(position)
            if not idx.isValid():
                return menu
            item = self.item_from_index(idx)
            if not item:
                return menu
            txid = txids[0]
            menu.add_action(translate("hist_list", "Details"), lambda: self.signals.open_tx_like.emit(txid))

            addr_URL = block_explorer_URL(self.config.network_config.mempool_url, "tx", txid)
            if addr_URL:
                menu.add_action(
                    translate("hist_list", "View on block explorer"),
                    lambda: webopen(addr_URL),
                    icon=read_QIcon("link.svg"),
                )
            menu.addSeparator()

            # addr_column_title = self._source_model.horizontalHeaderItem(
            #     self.Columns.LABEL
            # ).text()
            # addr_idx = idx.sibling(idx.row(), self.Columns.LABEL)
            self.add_copy_menu(menu, idx, include_columns_even_if_hidden=[self.Columns.TXID])
            # persistent = QPersistentModelIndex(addr_idx)
            # menu.add_action(
            #     translate("hist_list", "Edit {}").format(addr_column_title),
            #     lambda p=persistent: self.edit(QModelIndex(p)),
            # )
            # menu.add_action(translate("hist_list", "Request payment"), lambda: self.main_window.receive_at(txid))
            # if not is_multisig and not self.wallet.is_watching_only():
            #     menu.add_action(translate("hist_list", "Sign/verify message"), lambda: self.signals.sign_verify_message(txid))
            #     menu.add_action(translate("hist_list", "Encrypt/decrypt message"), lambda: self.signals.encrypt_message(txid))

        menu.add_action(
            translate("hist_list", "Copy as csv"),
            lambda: self.copyRowsToClipboardAsCSV([r.row() for r in selected]),
            icon=read_QIcon("csv-file.svg"),
        )

        menu.add_action(
            translate("hist_list", "Save as file"),
            lambda: self.export_raw_transactions(selected_items),
            icon=(self.style() or QStyle()).standardIcon(QStyle.StandardPixmap.SP_DialogSaveButton),
        )

        if not multi_select:
            idx = self.indexAt(position)
            if not idx.isValid():
                return menu
            item = self.item_from_index(idx)
            if not item:
                return menu
            txid = txids[0]

            wallet, tx_details = self._tx_dict[txid]
            tx_status = TxStatus.from_wallet(txid, wallet)
            if tx_status and tx_status.can_rbf():
                menu.addSeparator()
                menu.add_action(
                    translate("hist_list", "Edit with higher fee (RBF)"), lambda: self.edit_tx(tx_details)
                )

                menu.add_action(
                    translate("hist_list", "Try cancel transaction (RBF)"), lambda: self.cancel_tx(tx_details)
                )

        # run_hook('receive_menu', menu, txids, self.wallet)
        if viewport := self.viewport():
            menu.exec(viewport.mapToGlobal(position))

        return menu

    def edit_tx(self, tx_details: bdk.TransactionDetails) -> None:
        txinfos = ToolsTxUiInfo.from_tx(
            tx_details.transaction,
            FeeInfo.from_txdetails(tx_details),
            self.config.network,
            get_wallets(self.signals),
        )

        self.signals.open_tx_like.emit(txinfos)

    def cancel_tx(self, tx_details: bdk.TransactionDetails) -> None:
        txinfos = ToolsTxUiInfo.from_tx(
            tx_details.transaction,
            FeeInfo.from_txdetails(tx_details),
            self.config.network,
            get_wallets(self.signals),
        )

        wallet = get_wallet(self.wallet_id, self.signals)
        if not wallet:
            Message(
                self.tr("Cannot fetch wallet '{id}'. Please open the wallet first.").format(
                    id=self.wallet_id
                ),
                type=MessageType.Error,
            )
            return

        assert txinfos.spend_all_utxos, "Eeror in input selection for the cancel transaction"
        # it is ok to set amount=0, because  checked_max_amount=True
        amount = 0
        txinfos.recipients = [
            Recipient(
                wallet.get_address().address.as_string(),
                amount=amount,
                label=f"Cancel transaction {tx_details.txid}",
                checked_max_amount=True,
            )
        ]

        self.signals.open_tx_like.emit(txinfos)

    def export_raw_transactions(
        self, selected_items: Iterable[QStandardItem], folder: str | None = None
    ) -> None:
        if not folder:
            folder = QFileDialog.getExistingDirectory(None, "Select Folder")
            if not folder:
                logger.info("No file selected")
                return

        keys = [item.data(MyItemDataRole.ROLE_KEY) for item in selected_items]

        file_paths = self.drag_keys_to_file_paths(keys, save_directory=folder)

        logger.info(f"Saved {len(file_paths)} {self._source_model.drag_key} saved to {folder}")

    def get_edit_key_from_coordinate(self, row: int, col: int) -> Any:
        if col != self.Columns.LABEL:
            return None
        return self.get_role_data_from_coordinate(row, self.key_column, role=MyItemDataRole.ROLE_KEY)

    def on_edited(self, idx: QModelIndex, edit_key: str, *, text: str) -> None:
        txid = edit_key
        wallet, tx = self._tx_dict[txid]

        wallet.labels.set_tx_label(edit_key, text, timestamp="now")

        fulltxdetails = wallet.get_dict_fulltxdetail().get(txid)
        self.signals.wallet_signals[self.wallet_id].updated.emit(
            UpdateFilter(
                txids=[txid],
                addresses=(
                    [pythonutxo.address for pythonutxo in fulltxdetails.outputs.values() if pythonutxo]
                    if fulltxdetails
                    else []
                ),
                reason=UpdateFilterReason.UserInput,
            )
        )


class RefreshButton(QPushButton):
    def __init__(self, parent=None, height=20) -> None:
        super().__init__(parent)
        self.setText("")
        # Use the standard pixmap for the button icon
        self.setIconSize(QSize(height, height))  # Icon size can be adjusted as needed
        self.set_icon_allow_refresh()

    def set_icon_allow_refresh(self) -> None:
        icon = (self.style() or QStyle()).standardIcon(QStyle.StandardPixmap.SP_BrowserReload)
        self.setIcon(icon)

    def set_icon_is_syncing(self) -> None:

        icon = read_QIcon("status_waiting.svg")
        self.setIcon(icon)


class HistListWithToolbar(TreeViewWithToolbar):
    def __init__(self, hist_list: HistList, config: UserConfig, parent: QWidget | None = None) -> None:
        super().__init__(hist_list, config, parent=parent)
        self.hist_list = hist_list
        self.create_layout()

        self.sync_button = RefreshButton(height=QFontMetrics(self.balance_label.font()).height())
        self.sync_button.clicked.connect(self.hist_list.signals.request_manual_sync.emit)
        self.toolbar.insertWidget(0, self.sync_button)
        self.hist_list.signals.language_switch.connect(self.updateUi)
        self.hist_list.signals.wallet_signals[self.hist_list.wallet_id].updated.connect(self.updateUi)

    def updateUi(self) -> None:
        super().updateUi()
        if self.balance_label:
            balance_total = Satoshis(self.hist_list.balance, self.config.network)

            if self.hist_list.signals and not self.hist_list.address_domain:
                if self.hist_list.signals:
                    display_balance = (
                        self.hist_list.signals.wallet_signals[self.hist_list.wallet_id]
                        .get_display_balance.emit()
                        .get(self.hist_list.wallet_id)
                    )
                    if isinstance(display_balance, Balance):
                        balance_total = Satoshis(display_balance.total, self.config.network)

            self.balance_label.setText(balance_total.format_as_balance())

    def create_toolbar_with_menu(self, title) -> None:
        super().create_toolbar_with_menu(title=title)

        font = QFont()
        font.setPointSize(12)
        if self.balance_label:
            self.balance_label.setFont(font)

    def on_hide_toolbar(self) -> None:
        self.show_change = AddressTypeFilter.ALL  # type: AddressTypeFilter
        self.show_used = AddressUsageStateFilter.ALL  # type: AddressUsageStateFilter
        self.update()



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import sys

from PyQt6.QtCore import QRectF, QSize, QTimer, pyqtBoundSignal
from PyQt6.QtGui import QIcon, QPainter, QPaintEvent
from PyQt6.QtSvg import QSvgRenderer
from PyQt6.QtWidgets import QApplication, QMainWindow, QPushButton, QVBoxLayout, QWidget

from .util import icon_path


class SpinningButton(QPushButton):
    def __init__(
        self,
        text: str,
        enable_signal: pyqtBoundSignal | None = None,
        enabled_icon=QIcon(),
        spinning_svg_path=None,
        parent=None,
        timeout=20,
    ) -> None:
        super().__init__(text, parent)
        if spinning_svg_path is None:
            spinning_svg_path = icon_path("loader-icon.svg")
        self.svg_renderer = QSvgRenderer(spinning_svg_path)
        self.rotation_angle = 0
        self._icon_size = QSize(18, 18)  # Default icon size
        self.timer = QTimer(self)
        self.timeout_timer = QTimer(self)
        self.padding = 3
        self.timeout = timeout
        self.enabled_icon = enabled_icon
        self.setIcon(self.enabled_icon)

        self.clicked.connect(self.on_clicked)

        # Connect the external signal to the button's enable method
        if enable_signal:
            self.set_enable_signal(enable_signal)
        self.timeout_timer.timeout.connect(self.enable_button)

    def on_clicked(self) -> None:
        if not self.isEnabled():
            return
        self.setIcon(QIcon())
        self.start_spin()
        self.setDisabled(True)
        self.timeout_timer.start(self.timeout * 1000)

    def enable_button(self, *args, **kwargs) -> None:
        self.stop_spin()
        self.setIcon(self.enabled_icon)
        self.setEnabled(True)
        self.timeout_timer.stop()

    def set_enable_signal(self, enable_signal: pyqtBoundSignal) -> None:
        if enable_signal:
            enable_signal.connect(self.enable_button)

    def start_spin(self) -> None:
        # Timer to update rotation
        self.timer.timeout.connect(self.rotate_svg)
        self.timer.start(100)  # Update rotation every 100 ms

    def stop_spin(self) -> None:
        self.timer.stop()

    def setIconSize(self, size: QSize) -> None:
        if isinstance(size, QSize):
            self._icon_size = size
        else:
            raise TypeError("Size must be a QSize object")
        self.update()  # Redraw the button

    def iconSize(self) -> QSize:
        return self._icon_size

    def rotate_svg(self) -> None:
        self.rotation_angle = (self.rotation_angle + 10) % 360
        self.update()  # Trigger repaint

    def paintEvent(self, event: QPaintEvent | None) -> None:
        super().paintEvent(event)

        if self.timer.isActive():
            painter = QPainter(self)
            painter.setRenderHint(QPainter.RenderHint.Antialiasing)

            # Calculate the SVG's drawing rectangle
            icon_size = self.iconSize()
            icon_rect = QRectF(
                self.padding,
                (self.height() - icon_size.height()) / 2,
                icon_size.width(),
                icon_size.height(),
            )

            # Save painter's state to restore after drawing the icon
            painter.save()

            # Transform for rotation
            painter.translate(icon_rect.center())
            painter.rotate(self.rotation_angle)
            painter.translate(-icon_rect.center())

            # Draw SVG
            self.svg_renderer.render(painter, icon_rect)

            # Restore painter's state to draw the text
            painter.restore()

    def sizeHint(self) -> QSize:
        # Get the default size hint from the superclass
        default_size_hint = super().sizeHint()

        # Add icon width and padding to the width
        total_width = default_size_hint.width() + self._icon_size.width() + 2 * self.padding

        # Ensure the height is enough for the text and the icon
        total_height = max(default_size_hint.height(), self._icon_size.height())

        return QSize(total_width, total_height)


if __name__ == "__main__":

    class MainWindow(QMainWindow):
        def __init__(self) -> None:
            super(MainWindow, self).__init__()

            # Replace 'path/to/your.svg' with the path to your SVG file
            self.button = SpinningButton("Button Text", spinning_svg_path=icon_path("loader-icon.svg"))

            layout = QVBoxLayout()
            layout.addWidget(self.button)

            central_widget = QWidget()
            central_widget.setLayout(layout)
            self.setCentralWidget(central_widget)

    # Initialize the application
    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()

    # Run the application's event loop
    sys.exit(app.exec())
